"""Pyxis module for MS-related operations""";
import pyrap.tables
from pyrap.tables import table
import os.path
from astropy.io import fits as pyfits
import numpy as np

from Pyxis.ModSupport import *

import std
import _utils.casa_scripts
import im.argo

# register ourselves with Pyxis, and define the superglobals
register_pyxis_module();

v.define("MS","",
  """current measurement set""");
  
define("DDID",0,
  """current DATA_DESC_ID value""");
define("FIELD",0,
  """current FIELD value""");
define("IFRS","all",
  """interferometer subset""");
define("CHANRANGE",None,
  """channel range, as first,last[,step], or list of such tuples per DDID, or None for all""");
define("MS_TDL_Template",'ms_sel.msname=$MS ms_sel.ddid_index=$DDID ms_sel.field_index=$FIELD',
  """these options get passed to TDL scripts to specify an MS""");
  

lwimager = x.lwimager;
_flagms  	= x("flag-ms.py");
flagms          = _flagms.args("$MS ${-I <IFRS} $CHAN_OWLCAT");
downweigh_redundant = x("downweigh-redundant-baselines.py").args("$MS ${-I <IFRS}");
_aoflagger      = x.aoflagger;
addbitflagcol   = x.addbitflagcol.args("$MS");
wsrt_j2convert  = x.wsrt_j2convert.args("in=$MS");
mergems         = x("merge-ms.py");
taql		= x("taql");

define("PLOTVIS","CORRECTED_DATA:I",
  """passed to plot-ms to plot output visibilities. Set to None to skip plots.""");
define("PLOTMS_ARGS","",
  """extra plot-ms arguments""");
plotms = x("plot-ms.py").args("$MS $PLOTVIS $PLOTMS_ARGS ${-D <DDID} ${-F <FIELD} ${-I <IFRS} $CHAN_OWLCAT");

class _TableGuard (object):
  def __init__ (self,tab):
    self.tab = tab;
  def __enter__ (self):
    return self.tab;
  def __exit__ (self):
    self.tab.close();

def msw (msname="$MS",subtable=None):
  """Opens the MS or a subtable read-write, returns table object."""
  return ms(msname,subtable,write=True);

def ms (msname="$MS",subtable=None,write=False):
  """Opens the MS or a subtable (read-only by default), returns table object."""
  msname = interpolate_locals("msname");
  if not msname:
    raise ValueError("'msname' or global MS variable must be set");
  if subtable:
    msname = table(msname,ack=False).getkeyword(subtable);
  tab = table(msname,readonly=not write,ack=False);
  return tab;

def _filename (base,newext):
  while base and base[-1] == "/":
    base = base[:-1];
  return os.path.splitext(base)[0]+"."+newext;

def prep (msname="$MS"):
  """Prepares MS for use with MeqTrees: adds imaging columns, adds BITFLAG columns, copies current flags
  to 'legacy' flagset"""
  msname = interpolate_locals("msname");
  verify_antpos(msname,fix=True);
  add_imaging_columns(msname);
  info("adding bitflag column");
  x.addbitflagcol("$msname");
  info("copying FLAG to bitflag 'legacy'");
  _flagms("$msname -Y +L -f legacy -c");
  info("flagging INFs/NANs in data");
  _flagms("$msname --nan -f legacy --data-column DATA -x");
  
  
def add_imaging_columns (msname="$MS"):
  msname = interpolate_locals("msname");
  tab = msw(msname);
  if "MODEL_DATA" in tab.colnames() and "CHANNEL_SELECTION" in tab.getcolkeywords("MODEL_DATA"):
    tab.removecolkeyword('MODEL_DATA','CHANNEL_SELECTION');
  tab.close();
  import im.lwimager;
  if not im.lwimager.add_imaging_columns(msname):
    warn("Using pyrap to add imaging columns to $msname. Beware of https://github.com/ska-sa/lwimager/issues/3")
    pyrap.tables.addImagingColumns(msname);
    # if DATA column is not fixed shape, the MODEL_DATA and CORRECTED_DATA columns need to be initialized
    try:
      tab = ms(v.MS);
      x1 = tab.getcol("MODEL_DATA",0,1);
      x2 = tab.getcol("CORRECTED_DATA",0,1);
      info("MODEL_DATA shape",x1.shape[1:],"CORRECTED_DATA shape",x2.shape[1:]);
      return;
    except:
      info("will try to init MODEL_DATA and CORRECTED_DATA from DATA");
    copycol("DATA","MODEL_DATA",msname=msname);
    copycol("DATA","CORRECTED_DATA",msname=msname);
  
  
def delcols (*columns):  
  """Deletes the given columns in the MS""";
  tab = msw(v.MS);
  columns = [ col for col in columns if col in tab.colnames() ];
  info("deleting columns $columns");
  tab.removecols(columns);
  
def listcols (*columns):
  """With no arguments, lists all columns of the MS. With column names as arguments, lists shapes of specified columns""";
  tab = ms(v.MS);
  if not columns:
    info("columns are",*(tab.colnames()));
  else:
    for col in columns:
      coldata = tab.getcol(col);
      info("column $col shape",coldata.shape);
    info("MS has %d rows"%tab.nrows());
    
def verifycol (column):
  """Verifies that a column has data by reading it""";
  nddid = ms(v.MS,subtable="DATA_DESCRIPTION").nrows();
  info("$MS has $nddid DDIDs");
  tab0 = ms(v.MS);
  for ddid in range(nddid):
    tab = tab0.query("DATA_DESC_ID == %d"%ddid);
    coldata = tab.getcol(column);
    info("DDID $ddid column $column shape",coldata.shape);
  tab0.close()
      
document_globals(delcols,"MS");
document_globals(listcols,"MS");

def zerocol (column,ddid="$DDID",field="$FIELD",msname="$MS"):  
  """Fills the given column in the MS with zeroes""";
  column,msname,ddid,field = interpolate_locals("column msname ddid field");
  subtable = msw(msname).query(II("DATA_DESC_ID==$ddid && FIELD_ID==$field"));
  col = subtable.getcol(column);
  info("ddid $ddid field $field: will zero column $column of shape %s"%str(col.shape));
  col[...] = 0;
  subtable.putcol(column,col);
  subtable.close();
document_globals(zerocol,"MS DDID FIELD");
  
def copycol (fromcol="DATA",tocol="CORRECTED_DATA",rowchunk=500000,msname="$MS",to_ms="$msname",ddid=None,to_ddid=None):
  """Copies data from one column of MS to another.
  Copies 'rowchunk' rows at a time; decrease the default if you have low RAM.
  """;
  msname,destms,fromcol,tocol = interpolate_locals("msname to_ms fromcol tocol");
  if ddid is None:
    ddids = list(range(ms(msname,subtable="DATA_DESCRIPTION").nrows()));
    info("copying $msname $fromcol to $destms $tocol");
    info("$msname has %d DDIDs"%len(ddids));
    to_ddid = None;
  else:
    ddids = [ddid];
    if to_ddid is None:
      to_ddid = ddid;
    info("copying from $msname DDID $ddid $fromcol to $destms DDID $to_ddid $tocol");
  maintab0 = ms(msname);
  maintab1 = msw(destms);
  for ddid in ddids:
    tab0 = maintab0.query("DATA_DESC_ID == %d"%ddid);
    tab1 = maintab1.query("DATA_DESC_ID == %d"%(to_ddid if to_ddid is not None else ddid));
    nrows = tab0.nrows();
    info("DDID $ddid has $nrows rows");
    if tab1.nrows() != nrows:
      abort("table size mismatch: destination has %d rows"%tab1.nrows());
    for row0 in range(0,nrows,rowchunk):
      nr = min(rowchunk,nrows-row0);
      info("copying rows $row0 to %d"%(row0+nr-1));
      tab1.putcol(tocol,tab0.getcol(fromcol,row0,nr),row0,nr)
  for t in tab0,tab1,maintab0,maintab1:
    tab0.close()

def sumcols (fromcol1="DATA",fromcol2="MODEL_DATA",tocol="CORRECTED_DATA",rowchunk=500000,msname="$MS",to_ms="$msname",ddid=None,to_ddid=None):
  """Sums data from two columns of MS into a third.
  Copies 'rowchunk' rows at a time; decrease the default if you have low RAM.
  """;
  msname,destms,fromcol1,fromcol2,tocol = interpolate_locals("msname to_ms fromcol1 fromcol2 tocol");
  if ddid is None:
      ddids = list(range(ms(msname,subtable="DATA_DESCRIPTION").nrows()));
      info("copying $msname $fromcol1+$fromcol2 to $destms $tocol");
      info("$msname has %d DDIDs"%len(ddids));
      to_ddid = None;
  else:
      ddids = [ddid];
      if to_ddid is None:
        to_ddid = ddid;
      info("copying from $msname DDID $ddid $fromcol1+$fromcol2 to $destms DDID $to_ddid $tocol");
  maintab0 = ms(msname);
  maintab1 = msw(destms);
  for ddid in ddids:
    tab0 = maintab0.query("DATA_DESC_ID == %d"%ddid);
    tab1 = maintab1.query("DATA_DESC_ID == %d"%(to_ddid if to_ddid is not None else ddid));
    nrows = tab0.nrows();
    info("DDID $ddid has $nrows rows");
    if tab1.nrows() != nrows:
      abort("table size mismatch: destination has %d rows"%tab1.nrows());
    for row0 in range(0,nrows,rowchunk):
      nr = min(rowchunk,nrows-row0);
      info("copying rows $row0 to %d"%(row0+nr-1));
      tab1.putcol(tocol,tab0.getcol(fromcol1,row0,nr)+tab0.getcol(fromcol2,row0,nr),row0,nr)
  for t in tab0,tab1,maintab0,maintab1:
    tab0.close()

    
document_globals(copycol,"MS");

def verify_antpos (msname="$MS",fix=False,hemisphere=None):
  """Verifies antenna Y positions in MS. If Y coordinate convention is wrong, either fixes the positions (fix=True) or
  raises an error. hemisphere=-1 makes it assume that the observatory is in the Western hemisphere, hemisphere=1
  in the Eastern, or else tries to find observatory name using MS and pyrap.measure."""
  msname = interpolate_locals("msname");
  if not hemisphere:
    obs = ms(msname,"OBSERVATION").getcol("TELESCOPE_NAME")[0];
    info("observatory is $obs");
    try:
      import pyrap.measures
      hemisphere = 1 if pyrap.measures.measures().observatory(obs)['m0']['value'] > 0 else -1;
    except:
      traceback.print_exc();
      warn("$obs is unknown, or pyrap.measures is missing. Will not verify antenna positions.")
      return 
  info("antenna Y positions should be of sign %+d"%hemisphere);
  
  anttab = msw(msname,"ANTENNA");
  pos = anttab.getcol("POSITION");
  wrong = pos[:,1]<0 if hemisphere>0 else pos[:,1]>0;
  nw = sum(wrong);
  if nw:
    if not fix:
      abort("$msname/ANTENNA has $nw incorrect Y antenna positions. Check your coordinate conversions (from UVFITS?), or run pyxis ms.verify_antpos[fix=True]")
    pos[wrong,1] *= -1;
    anttab.putcol("POSITION",pos);
    info("$msname/ANTENNA: $nw incorrect antenna positions were adjusted (Y sign flipped)");
  else:
    info("$msname/ANTENNA: all antenna positions appear to have correct Y sign")

define('FIGURE_WIDTH',8,'width of plots, in inches');
define('FIGURE_HEIGHT',6,'height of plots, in inches');
define('FIGURE_DPI',100,'resolution of plots, in DPI');
  
def plot_uvcov (msname="$MS",width=None,height=None,dpi=None,save=None,select=None,limit=None,use_flags=False,**kw):
  """Makes uv-coverage plot
  'msname' is superglobal MS by default.
  If 'save' is given, saves figure to file.
  Use width/height/dpi to override figure settings.
  Any additional keyword arguments are passed to plot(). Try e.g. ms=.1 to change the marker size.
  Return value is maximum baseline length.
  """
  msname,save = interpolate_locals("msname save");
  tab = ms(msname);
  if select:
    tab = tab.query(select);
  uv = tab.getcol("UVW")[:,:2];
  if use_flags:
    flag = tab.getcol("FLAG_ROW")
    uv = uv[~flag,:]
  import pylab
  
  pylab.figure(figsize=(width or FIGURE_WIDTH,height or FIGURE_HEIGHT));
  pylab.plot(-uv[:,0],-uv[:,1],'.r',**kw);
  pylab.plot(uv[:,0],uv[:,1],'.b',**kw);
  mb = np.sqrt((uv**2).sum(1)).max();
  info("max baseline is %.3f km"%(mb*1e-3));
  if limit is not None:
    pylab.xlim(-limit,limit);
    pylab.ylim(-limit,limit);
  if save:
    pylab.savefig(save,dpi=(dpi or FIGURE_DPI));
    info("saved UV coverage plot to $save");
  else:
    pylab.show();
  return mb
  
document_globals(plot_uvcov,"MS FIGURE_*");  

def swapfields (f1,f2,msname="$MS"):
  """Swaps two fields in an MS"""
  msname = interpolate_locals("msname")
  info("swapping FIELDs $f1 and $f2 in $msname");
  field = msw(msname=msname,subtable="FIELD");
  for name in field.colnames():
    info("swapping column $name");
    col = field.getcol(name);
    # arrays needs to be swapped with copy, else confusion ensues
    if hasattr(col,'shape') and len(col.shape) > 1:
        col[f1],col[f2] = col[f2].copy(),col[f1].copy();
    else:
        col[f1],col[f2] = col[f2],col[f1]
    field.putcol(name,col);
  field.close();
  tab = msw();
  fcol = tab.getcol("FIELD_ID");
  r1 = (fcol==f1)
  r2 = (fcol==f2)
  fcol[r1] = f2
  fcol[r2] = f1
  tab.putcol("FIELD_ID",fcol);
  tab.close();


##
## ARCHIVE/UNARCHIVE FUNCTIONS
##
def_global("TARBALL_DIR",".","directory where tarballs of MSs are kept");

def load_tarball (msname="$MS"):
  """Unpacks fresh copy of MS from .tgz in ms.TARBALL_DIR"""; 
  msname = interpolate_locals("msname");
  if exists(msname):
    info("$msname exists, removing and unpacking fresh MS from tarball");
    x.sh("rm -fr $msname")
  else:
    info("$msname does not exist, unpacking fresh MS from tarball");
  x.sh("cd ${msname:DIR}; tar zxvf ${TARBALL_DIR}/${msname:FILE}.tgz");

def save_tarball (msname="$MS"):
  """Saves MS to .tgz in ms.TARBALL_DIR""";
  msname = interpolate_locals("msname");
  x.sh("cd ${msname:DIR}; tar zcvf ${TARBALL_DIR}/${msname:FILE}.tgz ${msname:FILE}");
                                
document_globals(load_tarball,"TARBALL_DIR");
document_globals(save_tarball,"TARBALL_DIR");


##
## MERGE/SPLIT/VIEW FUNCTIONS
##
def merge (output="merged.MS",options=""):
  """Merges the MSs given by MS_List into the output MS. Options are passed to merge-ms."""
  if not v.MS_List:
    abort("MS_List must be set before calling ms.merge()");
  output,options = interpolate_locals("output options");
  mergems("-f",options,output,*v.MS_List);
document_globals(merge,"MS_List");


def split_views (msname="$MS",output="${MS:DIR}/${MS:BASE}-%s.MS",column="OBSERVATION_ID",values=None):
  """Splits MS into views according to unique values of the column. If values is not specified,
  looks in the column for the set of unique values. Makes a set of reference MSs, using 'output' as
  a template.""";
  msname,output,column = interpolate_locals("msname output column");
  if not values:
    values = set(ms(msname).getcol(column));
  info("splitting $msname by $column (%s)"%" ".join(map(str,values)));
  for val in values:
    subset = output%val;
    taql("SELECT FROM $msname where $column == $val giving $subset",split_args=False);

    
def virtconcat (output="concat.MS",thorough=False,subtables=False):
  """Virtually concatenates the MSs given by MS_List into an output MS."""
  output,options = interpolate_locals("output options");
  auxfile = II("${output}.dat");
  if not v.MS_List:
    abort("MS_List must be set before calling ms.virtconcat()");
  if len(v.MS_List) != len(set([os.path.basename(x) for x in v.MS_List])):
    abort("ms.virtconcat: MSs to be concatenated need to have unique basenames. Please rename your MSs accordingly.");
  cmd = "";
  if thorough:
    cmd = """ms.open("%s", nomodify=False)\n"""%v.MS_List[0];
    for msl in v.MS_List[1:]:
      cmd += II("""ms.virtconcatenate("$msl","$auxfile",'1GHz','1arcsec')\n""");  
    cmd += II("""ms.close()\nos.remove('$auxfile')\n""");
  cmd += """if not ms.createmultims('$output',["%s"],
  [],
  True, # nomodify  
  False,# lock  
  $subtables): # copysubtables from first to all other members  
  os._exit(1);
ms.close();
"""%'","'.join(v.MS_List); 
  std.runcasapy(cmd);
  info("virtually concatenated %d inputs MSs into $output"%len(v.MS_List));

document_globals(virtconcat,"MS_List");

def concat (output="concat.MS",thorough=False,subtables=False,freqtol='1MHz',dirtol='1arcsec'):
  """Concatenates the MSs given by MS_List into an output MS."""
  output,freqtol,dirtol = interpolate_locals("output freqtol dirtol");
  info("concatenating",v.MS_List,"into $output");
  if not v.MS_List:
    abort("MS_List must be set before calling ms.concat()");
  if len(v.MS_List) != len(set([os.path.basename(name) for name in v.MS_List])):
    abort("ms.virtconcat: MSs to be concatenated need to have unique basenames. Please rename your MSs accordingly.");
  x.sh("cp -a "+v.MS_List[0]+" "+output);
  cmd = """ms.open("%s", nomodify=False)\n"""%output;
  for msl in v.MS_List[1:]:
    cmd += II("""ms.concatenate("$msl",freqtol='$freqtol',dirtol='$dirtol')\n""");  
  cmd += II("""ms.close()\n""");
  std.runcasapy(cmd);
  info("concatenated %d inputs MSs into $output"%len(v.MS_List));

document_globals(virtconcat,"MS_List");


##
## CONVERSION
##
def from_uvfits (fitsfile,msname="$MS"):
  """Converts UVFITS file into MS""";
  fitsfile,msname = interpolate_locals("fitsfile msname");
  if not msname:
    msname = fitsfile+".MS";
  std.runcasapy("""ms.fromfits(msfile='$msname',fitsfile='$fitsfile')""");
  verify_antpos(msname,fix=True);
  
def fixuvw (msname="$MS",fix=True,rowstep=100000):
  """Fixes the UVW column of an MS by recomputing it from scratch. If fix=False, only prints differences with current 
  UVW column at every rowstep-th row"""
  # these are parameters for the CASA script given by 
  msname = interpolate_locals("msname");
  write_uvw = fix;
  std.runcasapy(_utils.casa_scripts.fixuvw_casa,content="_utils.casa_scripts.fixuvw_casa");
  if fix:
    info("updated UVWs have been written to $msname");
  else:
    info("fix=False, updated UVWs not written out");


##
## RESAMPLING FUNCTIONS
##
def split_rebin (msname="$MS",output="$MSOUT",chan=None,time=None,field=None,spw=None,column="DATA"):
  """Splits and/or resamples MS in frequency with the given channel stepping size, and/or in time
  with the give time bin size (e.g. '5s'), and/or breaks out the specified fields and spws""";
  msname,output = interpolate_locals("msname output");
  args = ""
  def list2str (arg):
    return ",".join(map(str,stg)) if isinstance(arg,(list,tuple)) else str(arg);
  if chan:
    args += II(",width=[$chan]");
  if time:
    args += II(",timebin='$time'");
  if field:
    args += II(",field='%s'"%list2str(field));
  if spw:
    args += II(",spw='%s'"%list2str(spw));
  column = column.lower();
  std.runcasapy("""split(vis='$msname',outputvis='$output',datacolumn='$column'$args);""");
  


##
## INFO FUNCTIONS
##

def summary (msname="$MS"):
  msname = interpolate_locals("msname");
  std.runcasapy("listobs('$msname')");

##
## FLAGGING FUNCTIONS
##
def aoflagger (msname="$MS",strategy=None):
  """Runs AOFlagger with the specified strategy"""
  msname,strategy = interpolate_locals("msname strategy");
  if strategy:
    _aoflagger("-strategy $strategy $msname");
  else:
    _aoflagger(msname);


def flag_ifrs (msname="$MS",ifrs="",flagset="badifr"):
  """Flags specified baselines""";
  msname,ifrs,flagset = interpolate_locals("msname ifrs flagset");
  _flagms(msname,"-I $ifrs -f $flagset -c");


def_global("FLAG_CHANNELS_MULTIPLIER",1,"multiply channel numbers given to flag_channels() by N");
def_global("FLAG_TIMESLOTS_MULTIPLIER",1,"multiply timeslot numbers given to flag_timeslots() by N");

def flag_channels (msname="$MS",begin=0,end=0,ifrs="all",flagset="badchan"):
  """Flags specified channel range in the specified baselines""";
  msname,ifrs,flagset = interpolate_locals("msname ifrs flagset");
  begin *= FLAG_CHANNELS_MULTIPLIER;
  end *= FLAG_CHANNELS_MULTIPLIER;
  _flagms(msname,"-I $ifrs -L $begin~$end -f $flagset -c");
document_globals(flag_channels,"FLAG_CHANNELS_*");
  

def flag_timeslots (msname="$MS",begin=0,end=0,ifrs="all",flagset="badts"):
  """Flags specified channel range in the specified baselines""";
  msname,ifrs,flagset = interpolate_locals("msname ifrs flagset");
  begin *= FLAG_TIMESLOTS_MULTIPLIER;
  end *= FLAG_TIMESLOTS_MULTIPLIER;
  _flagms(msname,"-I $ifrs -T $begin~$end -f $flagset -c");
document_globals(flag_channels,"FLAG_TIMESLOTS_*");

  
  

###
### Various MS-related settings, setup automatically from MS variable
###


## current spwid and number of channels. Note that these are set automatically from the MS by the _msddid_Template below
define('SPWID',0,'currently selected spectral window, set automatically from DDID');
define('TOTAL_CHANNELS',0,'total number of channels in current spectral window');
define('SPW_CENTRE_MHZ',0,"centre frequency of current spectral window, MHz");
define('SPW_BANDWIDTH_MHZ',0,"bandwidth of current spectral window, MHz");

## whenever the MS or DDID changes, look up the corresponding info on channels and spectral windows 
_msddid = None;
def _msddid_accessed_Template ():
  global SPWID,TOTAL_CHANNELS,SPW_CENTRE_MHZ,SPW_BANDWIDTH_MHZ,_msddid;
  msddid = II("$MS:$DDID");
  if msddid != _msddid and II("$MS") and DDID is not None:
    _msddid = msddid;
    if not exists('$MS'):
      warn("$MS doesn't exist"); 
      return None;
    try:
      ddtab = ms(subtable="DATA_DESCRIPTION");
      if ddtab.nrows() < DDID+1:
        warn("No DDID $DDID in $MS");
        return None;
      SPWID = ddtab.getcol("SPECTRAL_WINDOW_ID",DDID,1)[0];
      spwtab = ms(subtable="SPECTRAL_WINDOW");
      TOTAL_CHANNELS = spwtab.getcol("NUM_CHAN",SPWID,1)[0];
#      SPW_CENTRE_MHZ = spwtab.getcol("REF_FREQUENCY",SPWID,1)[0]*1e-6;
      chans = spwtab.getcol("CHAN_FREQ",SPWID,1)[0];
      SPW_CENTRE_MHZ = (chans[0]+chans[-1])*1e-6/2;
      SPW_BANDWIDTH_MHZ = spwtab.getcol("TOTAL_BANDWIDTH",SPWID,1)[0]*1e-6;
      # make sure this is reevaluated
      _chanspec_Template();
      info("$MS ddid $DDID is spwid $SPWID, $TOTAL_CHANNELS channels, centred on $SPW_CENTRE_MHZ MHz, bandwidth $SPW_BANDWIDTH_MHZ MHz"); 
    except:
      warn("Error accessing $MS");
      traceback.print_exc();
      return None;
  return II("$MS:$DDID");

## whenever the channel range changes, setup strings for TDL & Owlcat channel selection (CHAN_TDL and CHAN_OWLCAT),
## and also CHANSTART,CHANSTEP,NUMCHANS
_chanspec = None;
def _chanspec_Template ():
  global CHAN_TDL,CHAN_OWLCAT,CHANSTART,CHANSTEP,NUMCHANS;
  chans = CHANRANGE;
  if isinstance(CHANRANGE,(list,tuple)) and type(CHANRANGE[0]) is not int:
    chans = CHANRANGE[DDID];
  # process channel specification 
  if chans is None:
    CHAN_OWLCAT = '';
    CHANSTART,CHANSTEP,NUMCHANS = 0,1,TOTAL_CHANNELS;
    CHAN_TDL = 'ms_sel.select_channels=0';
  else:
    if type(chans) is int:
      ch0,ch1,dch = chans,chans,1;
#      CHANSTART,CHANSTEP,NUMCHANS = chans,1,1;
    elif len(chans) == 1:
      ch0,ch1,dch = chans[0],chans[0],1;
#      CHANSTART,CHANSTEP,NUMCHANS = chans[0],1,1;
    elif len(chans) == 2:
      ch0,ch1,dch = chans[0],chans[1],1;
#      CHANSTART,CHANSTEP,NUMCHANS = chans[0],1,chans[1]-chans[0]+1;
    elif len(chans) == 3:
      ch0,ch1,dch = chans;
    CHANSTART,CHANSTEP,NUMCHANS = ch0,dch,((ch1-ch0)//dch+1);
    CHAN_OWLCAT = "-L %d~%d:%d"%(ch0,ch1,dch);
    CHAN_TDL = 'ms_sel.select_channels=1 ms_sel.ms_channel_start=%d ms_sel.ms_channel_end=%d ms_sel.ms_channel_step=%d'%\
               (ch0,ch1,dch);
  return CHANSTART,CHANSTEP,NUMCHANS;

def set_default_spectral_info():
  tab = ms(subtable='SPECTRAL_WINDOW')
  global CHANSTART,CHANSTEP,NUMCHANS,CHANRANGE
  NUMCHANS = tab.getcol('NUM_CHAN')[0]
  CHANSTART = 0
  CHANSTEP = 1
  CHANRANGE = CHANSTART,NUMCHANS-1,CHANSTEP


def create_empty_ms(msname="$MS", tel=None, pos=None, pos_type='casa', coords="itrf",
            synthesis=4, dtime=10, freq0="1.4GHz", dfreq="10MHz", lon_lat=None, **kw):
    """
Uses simms to create an empty measurement set. Requires
either an antenna table (CASA table) or a list of ITRF or ENU positions. 

msname: MS name
tel: Telescope name (This name must be in the CASA Database (check in me.obslist() in casapy)
     If its not there, then you will need to specify the telescope coordinates via "lon_lat"
pos: Antenna positions. This can either a CASA table or an ASCII file. 
     (see simms --help for more on using an ascii file)
pos_type: Antenna position type. Choices are (casa, ascii)
coords: This is only applicable if you are using an ASCII file. Choices are (itrf, enu)
synthesis: Synthesis time in hours
dtime: Integration time in seconds
freq0: Start frequency 
dfreq: Channel width
nbands: Number of frequency bands
**kw: extra keyword arguments.

A standard file should have the format: pos1 pos2 pos3* dish_diameter station
mount. NOTE: In the case of ENU, the 3rd position (up) is not essential and
may not be specified; indicate that your file doesn't have this dimension by
enebaling the --noup (-nu) option.
    """

    try:
        from simms import simms
    except ImportError:
        abort("Import simms failed. Please make sure you simms intalled.\n"
              "Find simms at github.com/SpheMakh/simms or install it by running"
              "  pip install simms")
    
    simms.create_empty_ms(msname=II(msname), tel=tel, pos=pos, pos_type=pos_type, 
                coords=coords, synthesis=synthesis, dtime=dtime, freq0=freq0, 
                dfreq=dfreq, lon_lat=lon_lat, **kw)



import im.argo

from Pyxis.ModSupport import *

register_pyxis_module(superglobals="MS LSM DESTDIR");
v.define("LSM","lsm.lsm.html","""current local sky model""");

# external tools  
define('IMAGER','lwimager','Imager to user. Default is lwimager.');
define('LWIMAGER_PATH','lwimager','path to lwimager binary. Default is to look in the system PATH.');
define('WSCLEAN_PATH','{im.WSCLEAN_PATH}','path to lwimager binary. Default is to look in the system PATH.');
define('MORESANE_PATH','{im.MORESANE_PATH}','path to lwimager binary. Default is to look in the system PATH.');

define('COLUMN','CORRECTED_DATA','default column to image');

# standard imaging options 
ifrs=""
npix=2048
cellsize="8arcsec"
mode="channel"
stokes="IQUV"
weight="briggs"
robust=0
wprojplanes=0
cachesize=4096
niter=1000
gain=.1
threshold=0
fixed=0

# rescale images by factor
flux_rescale=1

# use velocity rather than frequency
velocity = False;

no_weight_fov = False

# filenames for images 
define("BASENAME_IMAGE_Template","${OUTFILE}${-<IMAGER}","default base name for all image filenames below");
define("DIRTY_IMAGE_Template", "${BASENAME_IMAGE}.dirty.fits","output filename for dirty image");
define("PSF_IMAGE_Template", "${BASENAME_IMAGE}.psf.fits","output filename for psf image");
define("RESTORED_IMAGE_Template", "${BASENAME_IMAGE}.restored.fits","output filename for restored image");
define("RESIDUAL_IMAGE_Template", "${BASENAME_IMAGE}.residual.fits","output filename for deconvolution residuals");
define("MODEL_IMAGE_Template", "${BASENAME_IMAGE}.model.fits","output filename for deconvolution model");
define("FULLREST_IMAGE_Template", "${BASENAME_IMAGE}.fullrest.fits","output filename for LSM-restored image");
define("MASK_IMAGE_Template", "${BASENAME_IMAGE}.mask.fits","output filename for CLEAN mask");

# How to channelize the output image. 0 for average all, 1 to include all, 2 to average with a step of 2, etc.
# None means defer to 'imager' module options
define("IMAGE_CHANNELIZE",0,"image channels selection: 0 for all, 1 for per-channel cube")
# passed to tigger-restore when restoring models into images. Use e.g. "-b 45" for a 45" restoring beam.
define("RESTORING_OPTIONS","","extra options to tigger-restore for LSM-restoring")
# default clean algorithm
define("CLEAN_ALGORITHM","clark","CLEAN algorithm (clark, hogbom, csclean, etc.)")

def fits2casa (input,output):
    """Converts FITS image to CASA image."""
    im.argo.fits2casa(input,output)

def make_image (msname="$MS",column="$COLUMN",imager='$IMAGER',
                dirty=True,restore=False,restore_lsm=True,psf=False,
                dirty_image="$DIRTY_IMAGE",
                restored_image="$RESTORED_IMAGE",
                residual_image="$RESIDUAL_IMAGE",
                psf_image="$PSF_IMAGE",
                model_image="$MODEL_IMAGE",
                algorithm="$CLEAN_ALGORITHM",
                fullrest_image='${FULLREST_IMAGE}',
                restoring_options='${RESTORING_OPTIONS}',
                channelize=None,lsm="$LSM",**kw0):
    """Makes image(s) from MS. Set dirty and restore to True or False to make the appropriate images. You can also
    set either to a dict of options to be passed to the imager. If restore=True and restore_lsm is True and 'lsm' is set, 
    it will also make a full-restored image (i.e. will restore the LSM into the image) with tigger-restore. Use this when 
    deconvolving residual images. Note that RESTORING_OPTIONS are passed to tigger-restore.
  
    'channelize', if set, overrides the IMAGE_CHANNELIZE setting. If both are None, the options in the 'imager' module take effect.
  
    'algorithm' is the deconvolution algorithm to use (hogbom, clark, csclean, multiscale, entropy) 
  
    'dirty_image', etc. sets the image names, with defaults determined by the globals DIRTY_IMAGE, etc.
    """

    global IMAGER
    IMAGER =  II(imager)
    imager,msname,column,lsm,dirty_image,psf_image,restored_image,residual_image,model_image,algorithm,fullrest_image,restoring_options = \
interpolate_locals("imager msname column lsm dirty_image psf_image \
restored_image residual_image model_image algorithm fullrest_image restoring_options")
    

    makedir('$DESTDIR')
    if imager in ['lwimager','wsclean']:
        __import__('im.%s'%imager.lower());
        call_imager = eval( 'im.%s.make_image'%(imager.lower()) )
    else: 

        abort('Uknown imager: $imager')
    if MORESANE_PATH != '{im.MORESANE_PATH}':
       im.MORESANE_PATH = MORESANE_PATH
    if WSCLEAN_PATH != '{im.WSCLEAN_PATH}':
       im.WSCLEAN_PATH = WSCLEAN_PATH
    # make dict of imager arguments that have been specified globally or locally
    args_to_parse = 'npix weight robust stokes field no_weight_fov ifrs gain niter cachesize mode wprojplanes threshold cellsize'.split()
    kw = dict([ (arg,globals()[arg]) for arg in args_to_parse if arg in globals() and globals()[arg] is not None ])
    kw.update( pol=stokes,scale=im.argo.toDeg(cellsize),size='%d %d'%(npix,npix) )
    if imager == 'wsclean':
        kw.update(weight='%s %.2f'%(weight,robust) if weight=='briggs' else weight)
        if isinstance(threshold,str):
            kw.update(threshold=im.argo.toJy(threshold))
    kw.update([ (arg,kw[arg]) for arg in args_to_parse if arg in kw0 ])
    kw.update(**kw0)

    call_imager(msname,column=column,dirty=dirty,restore_lsm=restore_lsm,restore=restore,
                psf=psf,dirty_image=dirty_image,restored_image=restored_image,
                psf_image=psf_image,model_image=model_image,algorithm=algorithm,
                channelize=channelize,lsm=lsm,fullrest_image=fullrest_image,
                restoring_options=restoring_options,**kw)
    
document_globals(make_image,"*_IMAGE IMAGER COLUMN IMAGE_CHANNELIZE MS RESTORING_OPTIONS CLEAN_ALGORITHM ms.IFRS ms.DDID ms.FIELD ms.CHANRANGE")

def make_threshold_mask (input="$RESTORED_IMAGE",threshold=0,output="$MASK_IMAGE",high=1,low=0):
    """Makes a mask image by thresholding the input image at a given value. The output image is a copy of the input image,
    with pixel values of 'high' (1 by default) where input pixels are >= threshold, and 'low' (0 default) where pixels are <threshold.
    """
    input,output = interpolate_locals("input output");
    im.argo.make_threshold_mask(input=input,threshold=threshold,output=output,high=high,low=low)

document_globals(make_threshold_mask,"RESTORED_IMAGE MASK_IMAGE")

def predict_vis (msname="$MS",image="$MODEL_IMAGE",column="MODEL_DATA",channelize=None,
  copy=False,copyto="$COPY_IMAGE_TO",**kw0):
    """Converts image into predicted visibilities"""
    
    import im.lwimager

    msname,image,column,copyto = interpolate_locals("msname image column copyto")

    im.lwimager.predict_vis(msname=msname,image=image,column=column,channelize=channelize,
                            copy=copy,copyto=copyto,**kw0)

document_globals(predict_vis,"MS MODEL_IMAGE COPY_IMAGE_TO ms.IFRS ms.DDID ms.FIELD ms.CHANRANGE");

define("COPY_IMAGE_TO_Template", "${MS:BASE}.imagecopy.fits","container for image copy")
def make_empty_image (msname="$MS",image="${COPY_IMAGE_TO}",channelize=None,**kw0):
    msname,image = interpolate_locals("msname image")
    im.argo.make_empty_image(msname=msname,imagename=image,channelize=channelize,**kw0)

def make_psf (msname="$MS",**kw):
    """Makes an image of the PSF. All other arguments as per make_image()."""

    make_image(msname,dirty=False,psf=True,**kw)



"""Pyxis module for MS-related operations""";
from Pyxis.ModSupport import *

import argo
import ms
import im
import subprocess,glob

rm_fr = x.rm.args("-fr")
tigger_restore = x("tigger-restore")

# Keep wsclean gridding image? 
KEEP_GRIDDING_IMAGE = False        
# register ourselves with Pyxis and define the superglobals
register_pyxis_module(superglobals="MS LSM DESTDIR");

def STANDARD_IMAGER_OPTS_Template():
    """ Initialise standard imager options """
    global npix,cellsize,mode,stokes,weight,robust,niter,gain,threshold
    npix = im.npix
    cellsize = im.cellsize
    mode = im.mode
    stokes = im.stokes
    weight = im.weight
    robust = im.robust
    niter = im.niter
    gain = im.gain
    threshold = im.threshold

define('WSCLEAN_PATH','${im.WSCLEAN_PATH}','Path to WSCLEAN')
define('IMAGER','wsclean','Imager name')

# dict of known lwimager arguments, by version number
# this is to accommodate newer versions
_wsclean_known_args = {0:set('name predict size scale nwlayers minuvw maxuvw maxw pol '
                       'joinpolarizations multiscale multiscale-threshold-bias multiscale-scale-bias '
                       'cleanborder niter threshold gain mgain smallinversion '
                       'nosmallinversion smallpsf gridmode nonegative negative '
                       'stopnegative interval channelrange joinchannels '
                       'field weight natural mfsweighting superweight beamsize makepsf '
                       'imaginarypart datacolumn gkernelsize oversampling reorder no-reorder '
                       'addmodel addmodelapp savemodel wlimit'.split()),
                       1.4:set('channelsout mem absmem j'.split()),
                       1.5:set('fitbeam nofitbeam circularbeam ellipticalbeam beamshape tempdir '                                
                               'savegridding minuvw-m maxuvw-m'.split()),
                       1.6:set('dft-predict'),
                       1.7:set('moresane-ext casamask fitsmask mgain intervalsout ' 
                               'no-update-model-required saveweights'.split()),
                       1.8:set('moresane-arg moresane-sl'.split()),
                       1.9:set('fit-spectral-pol fit-spectral-log-pol deconvolution-channels'.split())
}

# whenever the path changes, find out new version number, and build new set of arguments
_wsclean_path_version = None,None;

def WSCLEAN_VERSION_Template (path='$WSCLEAN_PATH'):
    """ initilise imager arguments """

    # first check if wsclean is installed
    path = interpolate_locals('path')
    path = im.argo.findImager(path)
    if path is False:
        return -1

    global _wsclean_path_version,_wsclean_args
    if path != _wsclean_path_version[0]:
        _wsclean_path_version = path,wsclean_version()
        _wsclean_args = set()
        for version,args in _wsclean_known_args.items():
            if version <= _wsclean_path_version[1][0]:
                _wsclean_args.update(args)

    # the following options are not versions > 1.4 
    if _wsclean_path_version[1]>=1.4:
        for item in 'addmodel addmodelapp savemodel'.split():
            _wsclean_args.discard(item)
    return _wsclean_path_version[1]

def wsclean_version(path='${WSCLEAN_PATH}'):
    """ try to find wsclean version """

    path = interpolate_locals('path')
    std = subprocess.Popen([path,'-version'],stderr=subprocess.PIPE,stdout=subprocess.PIPE)
    if std.stderr.read():
        # if return error assume its version 0
        version = '0.0'
        tail = ""
    else:
        stdout = std.stdout.read().lower()
        version = stdout.split()
        try:
            ind = version.index('version')
            version = version[ind+1].split('-')[0]
            tail = version.split('-')[-1] if '-' in version else ""
        except ValueError:
            version = '0.0'
            tail = ""
    info('$path version is $version${-<tail}')

    if '.' in version:
        if version.startswith('1.9') or version.startswith('1.10') or \
            version.startswith('1.11') or version.startswith('1.12'):
            info("using wsclean 1.9 interface for $version")
            version = '1.9'
        try:
            version = list(map(int,version.split('.')))
        except ValueError: 
            version = 0,0
        vstr = '%d.' + '%d'*(len(version)-1)
        version = float(vstr%(tuple(version)))
    return version,tail

def _run(msname='$MS',clean=False,path='${im.WSCLEAN_PATH}',**kw):
    """ Run WSCLEAN """
    msname,path = interpolate_locals('msname path')
    path = argo.findImager(path,imager_name='WSCLEAN')
    if path is False:
        abort('Could not find WSCLEAN in system path, alises or at $path')

    # make dict of imager arguments that have been specified globally or locally
    args = dict([ (arg,globals()[arg.replace("-","_")]) for arg in _wsclean_args 
                    if arg.replace("-","_") in globals() and globals()[arg.replace("-","_")] is not None ]);
    args.update([ (arg,kw[arg]) for arg in _wsclean_args if arg in kw ])

    # map image size/resolution parameters
    csz = kw.get('cellsize',cellsize)
    np = kw.get('npix',npix)
    args['scale'] = csz if isinstance(csz,(int,float)) else argo.toDeg(csz)
    args['size']= '%d %d'%(np,np)

    # map weight parameters
    wgt = args['weight'];
    if wgt == "robust" or wgt == "briggs":
        args['weight'] = 'briggs %.2f'%(kw.get('robust',robust))

    # map threshold
    if isinstance(args['threshold'],str):
        args['threshold'] = im.argo.toJy(args['threshold'])
    
    # map clean parameter
    if not clean:
        args['niter'] = 0

    ms.FIELD is not None and args.setdefault('field',ms.FIELD)
    x.sh(argo.gen_run_cmd(path,args,suf='-',assign=' ',pos_args=[msname] if not isinstance(msname,(list,tuple)) else msname))
   
def make_image(msname='$MS',image_prefix='${im.BASENAME_IMAGE}',column='${im.COLUMN}',
               mslist=None,         # if given, overrieds msname
                path='${WSCLEAN_PATH}',
                imager='$IMAGER',
                restore=False,
                dirty=True,
                psf=False,
                restore_lsm=False,
                lsm='$LSM',
                algorithm='${im.CLEAN_ALGORITHM}',
                channelize=None,
                psf_image='${im.PSF_IMAGE}',
                dirty_image='${im.DIRTY_IMAGE}',
                model_image='${im.MODEL_IMAGE}',
                residual_image='${im.RESIDUAL_IMAGE}',
                restored_image='${im.RESTORED_IMAGE}',
                fullrest_image='${im.FULLREST_IMAGE}',
                restoring_options='${im.RESTORING_OPTIONS}',
                keep_component_images=False,
                **kw):
    """ run WSCLEAN """

    makedir('$DESTDIR')
    _imager = im.IMAGER
    im.IMAGER = II(imager)
    #Add algorithm label if required
    if im.DECONV_LABEL and restore:
        if isinstance(im.DECONV_LABEL,bool):
            if im.DECONV_LABEL:
                im.DECONV_LABEL = algorithm
    elif im.DECONV_LABEL is False:
        im.DECONV_LABEL = None

    path,msname,image_prefix,column,dirty_image,model_image,residual_image,restored_image,psf_image,channelize,\
      fullrest_image,restoring_options = \
      interpolate_locals('path msname image_prefix column dirty_image model_image residual_image '
                         'restored_image psf_image channelize fullrest_image restoring_options')

    # Check if WSCLEAN is where it is said to be
    path = argo.findImager(path,imager_name='WSCLEAN')
    
    # wsclean requires a WEIGHT_SPECTRUM column in the MS
    if wsclean_version()[0]<1.6:
        argo.addcol(msname,colname='WEIGHT_SPECTRUM',valuetype='float',init_with=1) 
    
    if 'datacolumn' not in list(kw.keys()):
        kw['datacolumn'] = column
     
    # Cater for moresane
    do_moresane = False
    if restore and algorithm.lower() in ['moresane','pymoresane']:
        kw['niter'] = 0
        kw['makepsf'] = True
        psf = True
        dirty = True
        if isinstance(restore,dict):
            kw0 = restore.copy()
        else: 
            kw0 = {}
        restore = False
        do_moresane = True
        from im import moresane
    else:
        if isinstance(restore,dict):
            kw.update(restore)
            restore = True
        elif not isinstance(restore,bool):
            restore = False

    kw['name'] = image_prefix    

    # Check channel selection options in kw
    if 'channelrange' in list(kw.keys()):
        if isinstance(kw['channelrange'],str):
            start,end = list(map(int,kw['channelrange'].split()))
        else:
            start,end = kw['channelrange']
    else:
        start,end = ms.CHANSTART,ms.CHANSTART+ms.NUMCHANS;
        # if multiple MSs are specified, adjust channel range
        if mslist:
            end = ms.TOTAL_CHANNELS*(len(mslist)-1) + end

    kw['channelrange'] = "%d %d"%(start,end);

    nr = 1 
    if not channelize:
        channelize = im.IMAGE_CHANNELIZE
    if channelize:
        nr = (end-start)//channelize
        kw['channelsout'] = nr
    if nr ==1:
        channelize=False
   
    if dirty: info("im.wsclean.make_image: making dirty image $dirty_image")
    if restore: info(" making restored image $restored_image\
                    (model is $model_image, residual is $residual_image)")
    
    if psf and not restore:
        kw['makepsf'] = True

    if 'pol' in list(kw.keys()):
        pol = kw['pol']
    elif 'stokes' in list(kw.keys()):
        pol = kw['pol'] = kw.pop('stokes')
    else:
        pol = stokes
        kw['pol'] = pol

    if ',' in pol:
        pol = pol.split(',')
    kw['clean'] = restore

    # also accepts list of MSs
    _run(mslist or msname,**kw)

    # delete gridding image unless user wants it
    if not KEEP_GRIDDING_IMAGE:
        rm_fr('$image_prefix-gridding.fits')

    # delete first-residual images
    first_residual_images = glob.glob(II("$image_prefix-*-first-residual.fits"))
    if first_residual_images:
        rm_fr(" ".join(first_residual_images))

    #TODO(sphe): always keep wsclean MFS images?
    # Combine images if needed
#    mfs = mode if 'mode' not in kw.keys() else kw['mode']
#    mfs = mode=='mfs'
#    abort(mfs,mode)

    def eval_list(vals):
        l = []
        for val in vals:
            l.append(II(val))
        return l

    def combine_pol(pol,image_prefix=None,mfs=False):
        dirtys = eval_list(['$image_prefix-%s-dirty.fits'%d for d in pol])
        if dirty:
            argo.combine_fits(dirtys,outname=dirty_image,ctype='STOKES',keep_old=keep_component_images)
        else:
            for item in dirtys:
                rm_fr(item)

        if restore:
            model = eval_list(['$image_prefix-%s-model.fits'%d for d in pol])
            argo.combine_fits(model,outname=model_image,ctype='STOKES',keep_old=keep_component_images)

            residual = eval_list(['$image_prefix-%s-residual.fits'%d for d in pol])
            argo.combine_fits(residual,outname=residual_image,ctype='STOKES',keep_old=keep_component_images)

            restored = eval_list(['$image_prefix-%s-image.fits'%d for d in pol])
            argo.combine_fits(restored,outname=restored_image,ctype='STOKES',keep_old=keep_component_images)

            if mfs:
                model_mfs = eval_list(['$image_prefix-MFS-%s-model.fits'%d for d in pol])
                argo.combine_fits(model_mfs,
                       outname=model_image.replace('.model.fits','-MFS.model.fits'),
                       ctype='STOKES',keep_old=keep_component_images)

                residual_mfs = eval_list(['$image_prefix-MFS-%s-residual.fits'%d for d in pol])
                argo.combine_fits(residual_mfs,
                       outname=residual_image.replace('.residual.fits','-MFS.residual.fits'),
                       ctype='STOKES',keep_old=keep_component_images)

                restored_mfs = eval_list(['$image_prefix-MFS-%s-image.fits'%d for d in pol])
                argo.combine_fits(restored_mfs,
                       outname=restored_image.replace('.restored.fits','-MFS.restored.fits'),
                       ctype='STOKES',keep_old=False)
        else:
            for fits in ['$image_prefix-%s-image.fits'%d for d in pol]:
                rm_fr(fits)

    if not channelize:
        if psf:
            x.mv('${image_prefix}-psf.fits $psf_image')
        elif restore:
            rm_fr('${image_prefix}-psf.fits')
        if len(pol)>1:
            combine_pol(pol,image_prefix)
        else:
            if dirty:
                x.mv('${image_prefix}-dirty.fits $dirty_image')
            else: 
                rm_fr('${image_prefix}-dirty.fits')

            if restore:
                 x.mv('${image_prefix}-model.fits $model_image')
                 x.mv('${image_prefix}-residual.fits $residual_image')
                 x.mv('${image_prefix}-image.fits $restored_image')
                # x.mv('${image_prefix}-psf.fits $psf_image')
            else:
                rm_fr('${image_prefix}-image.fits')
    else:
        # Combine component images from wsclean
        labels = ['%04d'%d for d in range(nr)]
        psfs = eval_list(['$image_prefix-%s-psf.fits'%d for d in labels])
        if psf: 
            argo.combine_fits(psfs,outname=II('$image_prefix.psf.fits'),ctype='FREQ',keep_old=keep_component_images)
        elif restore:
            for fits in psfs:
                rm_fr(fits)

        for i in pol:
            if len(pol) == 1:
               i = ''

            if i : 
                i = '-%s'%i
         
            dirtys = eval_list(['$image_prefix-%s$i-dirty.fits'%d for d in labels])
            if dirty:
                argo.combine_fits(dirtys,
                       outname=II('$image_prefix$i-dirty.fits') if i else dirty_image,
                       ctype='FREQ',keep_old=keep_component_images)
                if not restore:
                    xo.sh('rm -fr ${image_prefix}*image*.fits')
            else: 
                for fits in dirtys:
                    rm_fr(fits)

            if restore:
                model = eval_list(['$image_prefix-%s$i-model.fits'%d for d in labels])
                argo.combine_fits(model,outname=II('$image_prefix$i-model.fits') if i else model_image,
                       ctype='FREQ',keep_old=keep_component_images)

                residual = eval_list(['$image_prefix-%s$i-residual.fits'%d for d in labels])
                argo.combine_fits(residual,outname=II('$image_prefix$i-residual.fits') if i else residual_image,
                       ctype='FREQ',keep_old=keep_component_images)

                restored = eval_list(['$image_prefix-%s$i-image.fits'%d for d in labels])
                argo.combine_fits(restored,outname=II('$image_prefix$i-image.fits') if i else restored_image ,
                       ctype='FREQ',keep_old=keep_component_images)
                if len(pol)==1:
                    for old,new in zip([image_prefix+'-MFS-%s.fits'%img for img in 'model residual image'.split()],
                                       [model_image.replace('.model.fits','-MFS.model.fits'),
                                        residual_image.replace('.residual.fits','-MFS.residual.fits'),
                                        restored_image.replace('.restored.fits','-MFS.restored.fits')]):
                        #TODO(sphe) Should we always keep wsclean MFS images?
                        #if mfs:
                        x.mv('$old $new')
                        #else: 
                        #    rm_fr(old)

        if len(pol)>1:
            combine_pol(pol,image_prefix,mfs=True)

    if do_moresane:
        info(" im.moresane.deconv: making estored image $restored_image \
              model is $model_image, residual is $residual_image)")
        moresane.deconv(dirty_image,psf_image,model_image=model_image,
                           residual_image=residual_image,
                           restored_image=restored_image,**kw0)

    if restore or do_moresane:
        if lsm and restore_lsm:
            info("Restoring LSM into FULLREST_IMAGE=$fullrest_image");
            opts = restore_lsm if isinstance(restore_lsm,dict) else {};
            tigger_restore(restoring_options,"-f",
                           restored_image,lsm,
                           fullrest_image,
                           kwopt_to_command_line(**opts));
    
    im.IMAGER = _imager

document_globals(make_image,"im.*_IMAGE COLUMN im.IMAGE_CHANNELIZE MS im.RESTORING_OPTIONS im.CLEAN_ALGORITHM ms.IFRS ms.DDID ms.FIELD ms.CHANRANGE")



from Pyxis.ModSupport import *

import tempfile,glob,os.path,time,os

# register ourselves with Pyxis, and define the superglobals
register_pyxis_module();

v.define("OUTDIR","",
  """base output directory""");
v.define("SUFFIX_Template",'${spw<ms.DDID}',
  """suffix added to filenames, default is "-spwX""");
v.define("DESTDIR_Template",'${OUTDIR>/}plots-${MS:BASE>-}${spw<ms.DDID}',
  """destination directory for plots, images, etc.""");
v.define("OUTFILE_Template",'${DESTDIR>/}${MS:BASE>-}${SUFFIX>-}${s<STEP>-}${LABEL}',
  """base output filename for plots, images, etc.""");
v.define("STEP",1,
  """step counter, automatically incremented. Useful for decorating filenames.""")
v.define("LABEL","",
  """decorative label, mainly used for decorating filenames.""")
  
remove          = xo.rm.args("-fr");
copy            = x.cp.args("-a");
plotparms       = x("plot-parms.py").args("$PLOTPARMS_ARGS");
fitstool        = x("fitstool.py");

CASAPY_AUTO = xro.which("casapy").strip() or xro.which("casa").strip()

v.define("CASAPY_ZAPLOGS",True,"clean casapy*log and ipython*log files after successful execution of runcasapy")
v.define("CASAPY_ZAPLOGS_ALWAYS",False,"clean casapy*log and ipython*log files after any execution of runcasapy")
v.define("CASAPY_PATH",None,"explicit path to casapy/casa executable")

def runcasapy (command,content=None,zap=None,zap_always=None):
  """Runs the specified casapy command (which can be a multi-line script including newlines).
  'cleanlogs': cleanup casapy*log and ipython*log files after the run. If 0, never clean up.
    If 1, clean up if successful but not on errors. If 2, clean up always.
  'content': if not None, this is what's reported in the log. Otherwise the command itself is reported.
  """;
  command = interpolate_locals("command");
  zap = zap if zap is not None else CASAPY_ZAPLOGS;
  zap_always = zap_always if zap_always is not None else CASAPY_ZAPLOGS_ALWAYS; 
  
  casapy = CASAPY_PATH or CASAPY_AUTO
  if not casapy:
      abort("casapy/casa executable not found. Is CASA installed? You can also try setting std.CASAPY explicitly")

  # write command to script file
  tf = tempfile.NamedTemporaryFile(suffix=".py");
  # Call casapy from temp. directory to avoid conflicts when running parallel jobs
  td = tempfile.mkdtemp(dir='.')
  # we want get back to the working directory once casapy is launched
  cdir = os.path.realpath('.')

  command = "\nimport os\nos.chdir('%s')\n%s"%(cdir, command) 
  tf.write(command+"\nexit\n");
  tf.flush();
  tfname = tf.name;
  # run casapy
  if content:
    content = II(" ($content)");
  else:
    content = II(". Content:\n$command\n")
  info("Running casapy $tfname$content")
  t0 = time.time()
  retcode = x.sh("cd $td && $casapy --nologger --log2term -c $tfname")
  tf.close();
  # move content of temp. directory to working dir, then delete temp. directory
  x.sh("mv ${td}/* $cdir && rm -fr $td")
  # zap logs
  if zap_always or ( zap and not retcode ):
    logs = glob.glob("ipython-*.log") + glob.glob("casapy-*.log") + glob.glob("casa-*.log")
    logs = [ log for log in logs if os.path.getmtime(log) > t0 ];
    info("zapping CASA logfiles",*logs);
    for log in logs:
      if exists(log):
        os.unlink(log);
  if retcode:
    abort("casapy failed with return code %d. Check the logs for errors.");

document_globals(runcasapy,"CASAPY*");

def printpaths ():
  info("OUTDIR=$OUTDIR, DESTDIR=$DESTDIR, OUTFILE=$OUTFILE");




#Pyxis casa wrap
from Pyxis.ModSupport import *
import ms
import im
import subprocess
import tempfile
import im.argo
import pyrap.images as Images

# register ourselves with Pyxis and define the superglobals
register_pyxis_module(superglobals="MS LSM OUTDIR DESTDIR")

rm_fr = x.rm.args("-fr")
tigger_restore = x("tigger-restore")

define('CASA_PATH','${im.CASA_PATH}','Casa Path')
define('IMAGER','casa','Imager name')

v.define("LSM","lsm.lsm.html","""current local sky model""")

_casa_known_args = {4.10:set('vis imagename outlierfile field spw selectdata timerange uvrange antenna '
                              'scan observation mode nterms reffreq gridmode wprojplanes cfcache '
                              'painc niter gain threshold psfmode imagermode cyclefactor cyclespeedup '
                              'multiscale negcomponent smallscalebias interactive mask imsize cell '
                              'phasecenter restfreq stokes weighting robust npixels uvtaper outertaper '
                              'innertaper modelimage restoringbeam pbcor minpb usescratch allowchunk '
                              'async nchan start width interpolation chaniter outframe'.split()),
                    4.20:set(),
                    4.22:set('calready fitmachine mosweight noise scaletype veltype'.split()),
}

def STANDARD_IMAGING_OPTS_Template():
    global npix,cellsize,mode,stokes,weight,robust,niter,gain,threshold
    global wprojplanes,cachesize,ifrs,fixed,flux_rescale,velocity,no_weight_fov
    npix = im.npix
    cellsize = im.cellsize
    mode = im.mode
    stokes = im.stokes
    weight = im.weight
    robust = im.robust
    niter = im.niter
    gain = im.gain
    threshold = im.threshold
    wprojplanes = im.wprojplanes
    cachesize = im.cachesize
    ifrs = im.ifrs
    fixed = im.fixed
    # rescale images by factor
    flux_rescale= im.flux_rescale
    # use velocity rather than frequency
    velocity = im.velocity
    no_weight_fov = im.no_weight_fov


# whenever the path changes, find out new version number, and build new set of arguments
_casa_path_version = None,None;
def CASA_VERSION_Template (path='$CASA_PATH'):
    path = interpolate_locals('path')
    path = im.argo.findImager(path)
    if not path:
        return -1
    global _casa_path_version,_casa_args
    if path != _casa_path_version[0]:
        _casa_path_version = path,casa_version(path)
        _casa_args = set()
        for version,args in _casa_known_args.items():
            if version <= _casa_path_version[1][0]:
                _casa_args.update(args)
    return _casa_path_version[1]


def casa_version(path='$CASA_PATH'):
    """ try to find casa version """

    path = interpolate_locals('path')
    std = subprocess.Popen([path,'--help','--log2term','--nologger','--nogui','--help','-c','quit'],
                            stderr=subprocess.PIPE,stdout=subprocess.PIPE)
    if std.stderr.read():
        version = '4.10' # start support from casa 4.10
        tail = ""
    else:
        stdout = std.stdout.read().lower()
        version = stdout.split()
        ind = version.index('version')
        version = version[ind+1].split('-')[0]
        tail = version.split('-')[-1] if '-' in version else ""

    info('$path version is $version${-<tail}')

    if '.' in version:
        try:
            version = list(map(int,version.split('.')))
        except ValueError: 
            version = 4,1,0
        vstr = '%d.' + '%d'*(len(version)-1)
        version = float(vstr%(tuple(version)))
    return version,tail


def _run(path='${im.CASA_PATH}',clean=False,makepsf=False,**kw):
    """ runs casa's clean task """
    
    path = interpolate_locals('path')
    # map some options to casapy equivalents
    global cell,imsize,weighting
    cell,imsize,weighting = cellsize,npix,weight

    # make dict of imager arguments that have been specified globally or locally
    args = dict([ (arg,globals()[arg]) for arg in _casa_args if arg in globals() and globals()[arg] is not None ]);
    args.update([ (arg,kw[arg]) for arg in _casa_args if arg in kw ])
    # add ifrs, spwid and field arguments
    ms.IFRS is not None and args.setdefault('ifrs',ms.IFRS)
    ms.DDID is not None and args.setdefault('spw',ms.DDID)
    ms.FIELD is not None and args.setdefault('field',ms.FIELD)
    
    if args.get("wprojplanes",0):
        args["gridmode"] = "widefield"
    

    # have an IFR subset? Parse that too
    msname,ifrs = args['vis'],args.pop('ifrs',None)
    if ifrs and ifrs.lower() != "all":
        import Meow.IfrSet
        subset = Meow.IfrSet.from_ms(msname).subset(ifrs).taql_string()
        args['selectdata'] = "(%s)&&(%s)"%(args['selectdata'],subset) if 'selectdata' in args else subset
    
    # casapy likes these options as strings 
    for item in 'field spw'.split():
        if not isinstance(args[item],str):
            args[item] = str(args[item])

    im.argo.icasa('clean',**args)
    imagename = kw['imagename']
    # convert casa images to fits files.
    mult = []
    if clean:
        model = kw['model']
        residual = kw['residual']
        restored = kw['restored']
        for img,fitsim in zip('model residual image'.split(),[model,residual,restored]):
            mult.append({'imagename':'%s.%s'%(imagename,img),'fitsimage':fitsim})
    else:
        mult.append({'imagename':'%s.%s'%(imagename,'image'),'fitsimage':kw['dirty']})
    if makepsf: 
        mult.append({'imagename':'%s.%s'%(imagename,'psf'),'fitsimage':kw['psf']})
    if mult:
        velo = kw.get("velocity") or velocity
        fs = kw.get("flux_rescale") or flux_rescale
        for pair in mult:
            casaim, fitsim = pair["imagename"],pair["fitsimage"]
            _im = Images.image(casaim)
            if fs!=1:
                _im.putdata(fs*_im.getdata())
            if os.path.exists(casaim):
                _im.tofits(fitsim, overwrite=True, velocity=velo)
            else:
                abort("Cannot find images. Something went wrong when running CASAPY clean task. Please check logs")
        

#        im.argo.icasa('exportfits',mult=mult,overwrite=True)
    # delete casa images
    for image in ['$imagename.%s'%s for s in 'model residual image flux psf'.split()]:
        if os.path.exists(II(image)):
            rm_fr(image)


def make_image (msname="$MS",column="${im.COLUMN}",imager='$IMAGER',
                dirty=True,restore=False,restore_lsm=True,psf=False,
                dirty_image="${im.DIRTY_IMAGE}",
                model_image="${im.MODEL_IMAGE}",
                restored_image="${im.RESTORED_IMAGE}",
                residual_image="${im.RESIDUAL_IMAGE}",
                psf_image="${im.PSF_IMAGE}",
                algorithm="${im.CLEAN_ALGORITHM}",
                fullrest_image="${im.FULLREST_IMAGE}",
                restoring_options="${im.RESTORING_OPTIONS}",
                channelize=None,lsm="$LSM",**kw0):
    """ run casa imager """

    _imager = im.IMAGER
    im.IMAGER = II(imager)
    #Add algorithm label if required
    if im.DECONV_LABEL and restore: 
        if isinstance(im.DECONV_LABEL,bool):
            if im.DECONV_LABEL:
                im.DECONV_LABEL = algorithm
    elif im.DECONV_LABEL is False:
        im.DECONV_LABEL = None
     
    do_moresane=False
    if algorithm.lower() in ['moresane','pymoresane']: 
        do_moresane = True
        from im import moresane
    imager,msname,column,lsm,dirty_image,psf_image,restored_image,residual_image,\
model_image,algorithm,fullrest_image,restoring_options = \
interpolate_locals("imager msname column lsm dirty_image psf_image restored_image "
                   "residual_image model_image algorithm fullrest_image restoring_options")

    makedir('$DESTDIR')

    if restore and column != "CORRECTED_DATA":
        abort("Due to imager limitations, restored images can only be made from the CORRECTED_DATA column.")

    # setup imager options
    kw0.update(dict(chanstart=ms.CHANSTART,chanstep=ms.CHANSTEP,nchan=ms.NUMCHANS));
    if 'nchan' not in kw0 or 'start' not in kw0:
        if channelize is None:
            channelize = im.IMAGE_CHANNELIZE;
        if channelize == 0:
            kw0.update(nchan=1,width=ms.NUMCHANS);
        elif channelize > 0:
            kw0.update(nchan=ms.NUMCHANS//channelize,start=ms.CHANSTART,width=channelize);

    kw0.update(vis=msname,imagename=im.BASENAME_IMAGE)
    def make_dirty():
        info("im.casa.make_image: making dirty image $dirty_image")
        kw = kw0.copy()
        kw['niter'] = 0
        _run(dirty=dirty_image,makepsf=psf,psf=psf_image,**kw)
    if dirty:
        make_dirty()

    if do_moresane and restore:
        if np.logical_or(not dirty,not psf): 
            psf = True
            make_dirty()

        opts = restore if isinstance(restore,dict) else {}
        info(" making restored image $restored_image\
                    (model is $model_image, residual is $residual_image)")

        moresane.deconv(dirty_image,psf_image,model_image=model_image,
                           residual_image=residual_image,restored_image=restored_image,**opts)
    elif restore:
        info(" making restored image $restored_image\
              (model is $model_image, residual is $residual_image)")
        kw = kw0.copy()
        kw['psfmode'] = algorithm if algorithm in 'clark clarkstokes hogbom' else 'clark'
        if isinstance(restore,dict):
            kw.update(restore)
        _run(model=model_image,residual=residual_image,restored=restored_image,makepsf=psf,psf=psf_image,clean=True,**kw)

    if restore:
        if lsm and restore_lsm:
            info("Restoring LSM into FULLREST_IMAGE=$fullrest_image")
            opts = restore_lsm if isinstance(restore_lsm,dict) else {}
            tigger_restore(restoring_options,"-f",restored_image,lsm,fullrest_image,kwopt_to_command_line(**opts))

    im.IMAGER = _imager

document_globals(make_image,"im.*_IMAGE COLUMN im.IMAGE_CHANNELIZE MS im.RESTORING_OPTIONS im.CLEAN_ALGORITHM ms.IFRS ms.DDID ms.FIELD ms.CHANRANGE")



"""Pyxis module fo0r MS-related operations""";
from Pyxis.ModSupport import *

import pyrap.images
import os
import subprocess
from astropy.io import fits as pyfits
import argo

import ms
import std
import im
import numpy 
# register ourselves with Pyxis and define the superglobals
register_pyxis_module(superglobals="MS LSM DESTDIR");

v.define("LSM","lsm.lsm.html","""current local sky model""");
  
# external tools  
define('LWIMAGER_PATH','lwimager','path to lwimager binary. Default is to look in the system PATH.');

define('IMAGER','lwimager','Imager name')

# dict of known lwimager arguments, by version number
# this is to accommodate newer versions
_lwimager_known_args = {
  0:set(("ms spwid field prior image model restored residual data mode filter nscales weight weight_fov noise robust wprojplanes padding "+
    "cachesize stokes nfacets npix cellsize phasecenter field spwid chanmode nchan chanstart chanstep img_nchan img_chanstart img_chanstep "+
    "select operation niter gain threshold targetflux sigma fixed constrainflux prefervelocity mask maskblc masktrc uservector maskvalue").split(" ")),
  1003001:set(["fillmodel"])
};

# whenever the path changes, find out new version number, and build new set of arguments
_lwimager_path_version = None,None;
def LWIMAGER_VERSION_Template ():
  global _lwimager_path_version,_lwimager_args;
  if LWIMAGER_PATH != _lwimager_path_version[0]:
    _lwimager_path_version = LWIMAGER_PATH,lwimager_version();
    _lwimager_args = set();
    for version,args in _lwimager_known_args.items():
      if version <= _lwimager_path_version[1][0]:
        _lwimager_args.update(args); 
  return _lwimager_path_version[1];

rm_fr = x.rm.args("-fr");
tigger_restore = x("tigger-restore");

imagecalc = x("imagecalc");

def STANDARD_IMAGING_OPTS_Template():
    global npix,cellsize,mode,stokes,weight,robust,niter,gain,threshold
    global wprojplanes,cachesize,ifrs,fixed,flux_rescale,velocity,no_weight_fov
    npix = im.npix
    cellsize = im.cellsize
    mode = im.mode
    stokes = im.stokes
    weight = im.weight
    robust = im.robust
    niter = im.niter
    gain = im.gain
    threshold = im.threshold
    wprojplanes = im.wprojplanes
    cachesize = im.cachesize
    ifrs = im.ifrs
    fixed = im.fixed
    # rescale images by factor
    flux_rescale= im.flux_rescale
    # use velocity rather than frequency
    velocity = im.velocity
    no_weight_fov = im.no_weight_fov

# known lwimager args -- these will be passed from keywords
_fileargs = set("image model restored residual".split(" ")); 


def add_imaging_columns (msname="$MS"):
  """Uses lwimager to insrt MODEL_DATA and CORRECTED_DATA columns""";
  msname = interpolate_locals("msname");
  if LWIMAGER_VERSION[0] >= 1003002:
    info("using $LWIMAGER_PATH to add imaging columns to $msname");
    x.sh("$LWIMAGER_PATH ms=$msname operation=empty fillmodel=1 image=$msname-dummy-pyxis.img");
    x.sh("rm -fr $msname-dummy-pyxis.img")
    return True;
  else:
    warn("lwimager >= 1.3.2 needed to add imaging columns to an MS");
    return None;


def _run (convert_output_to_fits=True,lwimager_path="$LWIMAGER_PATH",**kw):
  # look up lwimager
  lwimager_path = interpolate_locals("lwimager_path");
  lwimager_path = argo.findImager(lwimager_path)
  # make dict of imager arguments that have been specified globally or locally
  args = dict([ (arg,globals()[arg]) for arg in _lwimager_args if arg in globals() and globals()[arg] is not None ]);
  args.update([ (arg,kw[arg]) for arg in _lwimager_args if arg in kw ]);
  if no_weight_fov:
    args.pop('weight_fov',0);
  # add ifrs, spwid and field arguments
  ms.IFRS is not None and args.setdefault('ifrs',ms.IFRS);
  ms.DDID is not None and args.setdefault('spwid',ms.DDID);
  ms.FIELD is not None and args.setdefault('field',ms.FIELD);
  # have an IFR subset? Parse that too
  msname,ifrs = kw['ms'],args.pop('ifrs',None);
  if ifrs and ifrs.lower() != "all":
    import Meow.IfrSet
    subset = Meow.IfrSet.from_ms(msname).subset(ifrs).taql_string();
    args['select'] = "(%s)&&(%s)"%(args['select'],subset) if 'select' in args else subset;
  # make image names
  fitsfiles = {};
  for arg in _fileargs:
    if arg in args:
      if not args[arg].endswith(".img"):
        fitsfiles[arg] = args[arg];
        args[arg] = args[arg]+".img";
  # run the imager
  x.time("$lwimager_path",**args);
  # convert CASA images to FITS
  if convert_output_to_fits:
    fs = kw.get('flux_rescale') or flux_rescale;
    velo = kw.get('velocity') or velocity;
    for arg in _fileargs:
      if arg in fitsfiles:
        _im = pyrap.images.image(args[arg]);
        if fs and fs != 1:
          _im.putdata(fs*_im.getdata());
        _im.tofits(fitsfiles[arg],overwrite=True,velocity=velo);  
        subprocess.call("rm -fr "+args[arg],shell=True);


def lwimager_version (path="$LWIMAGER_PATH"):
  """Determines lwimager version, returns tuple of xxxyyyzzz,tail, where 
  xxx,yyy,zzz is major, minor, patch numbers, while tail is the rest of 
  the version string. Takes into account idiosyncarsies of older lwimager
  version strings, so for example:
            1003001,"" (for 1.3.1)
            1002001,"20120223-OMS" (for 1.2.1-20120223-OMS)
            1003000,"20130816-OMS" (for 20130816-OMS which is really 1.3.0)
  """
  path = interpolate_locals("path");
  try:
    vstr = subprocess.Popen([path,"--version"],stderr=subprocess.PIPE).stderr.read().strip().split()[-1];
  except:
    return 0,"";
  if '.' in vstr:
    major,minor,patch = vstr.split('.')
    patch,tail = patch.split("-",1) if "-" in patch else (patch,"");
  else:
    major,minor,patch,tail = 1,3,0,vstr;
  try:
    major,minor,patch = list(map(int,[major,minor,patch]));
  except:
    major,minor,patch,tail = 0,0,0,vstr;
  info("$path version is $major.$minor.$patch${-<tail}")
  return major*1000000+minor*1000+patch,tail;


def make_image (msname="$MS",column="${im.COLUMN}",imager='$IMAGER',
                dirty=True,restore=False,restore_lsm=True,psf=False,
                dirty_image="${im.DIRTY_IMAGE}",
                restored_image="${im.RESTORED_IMAGE}",
                residual_image="${im.RESIDUAL_IMAGE}",
                psf_image="${im.PSF_IMAGE}",
                model_image="${im.MODEL_IMAGE}",
                fullrest_image="${im.FULLREST_IMAGE}",
                restoring_options="${im.RESTORING_OPTIONS}",
                algorithm="${im.CLEAN_ALGORITHM}",
                channelize=None,lsm="$LSM",
                double_psf=None,**kw0):
  """Makes image(s) from MS. Set dirty and restore to True or False to make the appropriate images. You can also
  set either to a dict of options to be passed to the imager. If restore=True and restore_lsm is True and 'lsm' is set, 
  it will also make a full-restored image (i.e. will restore the LSM into the image) with tigger-restore. Use this when 
  deconvolving residual images. Note that RESTORING_OPTIONS are passed to tigger-restore.
  
  'channelize', if set, overrides the IMAGE_CHANNELIZE setting. If both are None, the options in the 'imager' module take effect.
  
  'algorithm' is the deconvolution algorithm to use (hogbom, clark, csclean, multiscale, entropy) 
  
  'dirty_image', etc. sets the image names, with defaults determined by the globals DIRTY_IMAGE, etc.
  """;
  
  _imager = im.IMAGER
  im.IMAGER = II(imager)
  # retain lwimager label for dirty maps and psf_maps
  #Add algorithm label if required
  if im.DECONV_LABEL and restore:
    if isinstance(im.DECONV_LABEL,bool):
      if im.DECONV_LABEL:
        im.DECONV_LABEL = algorithm
  elif im.DECONV_LABEL is False:
    im.DECONV_LABEL = None

  do_moresane = False
  if algorithm.lower() in ['moresane','pymoresane']:
      from im import moresane
      do_moresane = True

  imager,msname,column,lsm,dirty_image,psf_image,restored_image,residual_image,model_image,algorithm,\
     fullrest_image,restoring_options,double_psf = \
     interpolate_locals("imager msname column lsm dirty_image psf_image restored_image "
                        "residual_image model_image algorithm fullrest_image restoring_options double_psf");
  makedir('$DESTDIR');
  if restore and column != "CORRECTED_DATA":
    abort("Due to imager limitations, restored images can only be made from the CORRECTED_DATA column.");

  # setup imager options
  kw0.update(dict(chanstart=ms.CHANSTART,chanstep=ms.CHANSTEP,nchan=ms.NUMCHANS));
  if 'img_nchan' not in kw0 or 'img_chanstart' not in kw0:
    if channelize is None:
      channelize = im.IMAGE_CHANNELIZE;
    if channelize == 0:
      kw0.update(img_nchan=1,img_chanstart=ms.CHANSTART,img_chanstep=ms.NUMCHANS);
    elif channelize > 0:
      kw0.update(img_nchan=ms.NUMCHANS//channelize,img_chanstart=ms.CHANSTART,img_chanstep=channelize);
  
  kw0.update(ms=msname,data=column);

  def make_dirty(**kw1):
    info("im.lwimager.make_image: making dirty image $dirty_image");
    kw = kw0.copy();
    if type(dirty) is dict:
      kw.update(dirty);
    kw.update(kw1)
    kw['operation'] = 'image';
    _run(image=dirty_image,**kw);

  if dirty: make_dirty()

  def make_psf(**kw1):
    info("im.lwimager.make_image: making PSF image $psf_image");
    kw = kw0.copy();
    if type(psf) is dict:
      kw.update(psf);
    kw['operation'] = 'image';
    kw['data'] = 'psf';
    kw['stokes'] = "I";
    kw.update(kw1)
    _run(image=psf_image,**kw);
  if psf: make_psf()

  if do_moresane and restore:
    # Moresane does better with a double sized PSF
    double_psf = double_psf or im.DOUBLE_PSF
    if double_psf: 
        _npix = int(npix)*2 if 'npix' not in list(kw0.keys()) else (kw0['npix'])*2
        make_psf(npix=_npix)
    elif not psf: 
         make_psf()
    if not dirty: make_dirty()  
    opts = restore if isinstance(restore,dict) else {}
    moresane.deconv(dirty_image,psf_image,model_image=model_image,
                       residual_image=residual_image,restored_image=restored_image,**opts)
  elif restore:
    info("im.lwimager.make_image: making restored image $restored_image");
    info("                   (model is $model_image, residual is $residual_image)");
    kw = kw0.copy();
    if type(restore) is dict:
      kw.update(restore);
    kw.setdefault("operation",algorithm or "clark");
    temp_images = [];
    ## if fixed model was specified as a fits image, convert to CASA image
    if kw.pop('fixed',None):
      kw['fixed'] = 1;
      if not os.path.exists(model_image):
        warn("fixed=1 (use prior model) specified, but $model_image does not exist, ignoring"); 
      elif not os.path.isdir(model_image):
        info("converting prior model $model_image into CASA image");
        modimg = model_image+".img";
        temp_images.append(modimg);
        argo.fits2casa(model_image,modimg);
        model_image = modimg;
    ## if mask was specified as a fits image, convert to CASA image
    mask = kw.get("mask");
    if mask and not isinstance(mask,str):
      kw['mask'] = mask = im.MASK_IMAGE; 
    imgmask = None;
    if mask and os.path.exists(mask) and not os.path.isdir(mask):
      info("converting clean mask $mask into CASA image");
      kw['mask'] = imgmask = mask+".img";
      argo.fits2casa(mask,imgmask);
      temp_images.append(imgmask);
    ## run the imager
    _run(restored=restored_image,model=model_image,residual=residual_image,**kw)
    ## delete CASA temp images if created above
    for img in temp_images:
      rm_fr(img);
  if restore:
    if lsm and restore_lsm:
      info("Restoring LSM into FULLREST_IMAGE=$fullrest_image");
      opts = restore_lsm if isinstance(restore_lsm,dict) else {};
      tigger_restore(restoring_options,"-f",restored_image,lsm,fullrest_image,kwopt_to_command_line(**opts));

  im.IMAGER = _imager
      
document_globals(make_image,"im.*_IMAGE COLUMN im.IMAGE_CHANNELIZE MS im.RESTORING_OPTIONS im.CLEAN_ALGORITHM ms.IFRS ms.DDID ms.FIELD ms.CHANRANGE");      

def_global("PREDICT_CHANCHUNK",None,"use a maximum chunk size (in channels) when predicting image cubes")
def_global("PREDICT_KEEPIMAGES",False,"do not delete CASA images made while predicting image cubes")

def predict_vis (msname="$MS",image="${im.MODEL_IMAGE}",column="MODEL_DATA",
  channelize=None,
  chanchunk=None,
  copy=False,copyto="$COPY_IMAGE_TO",**kw0):
  """Converts image into predicted visibilities"""
  msname,image,column,copyto = interpolate_locals("msname image column copyto")
  chanchunk = chanchunk or PREDICT_CHANCHUNK
  
  if LWIMAGER_VERSION[0] in (1003000,1003001):
    abort("lwimager 1.3.%d cannot be used to predict visibilities. Try lwimager-1.2, or upgrade to 1.3.2 or higher"%(LWIMAGER_VERSION[0]%1000))
  
#  # copy data into template, if specified
#  if copy:
#    ff0 = pyfits.open(image);
#    info("copying image data from $image to $copyto");
#    npix = ff0[0].header['NAXIS1'];
#    cell = "%fdeg"%abs(ff0[0].header['CDELT1']);
#    make_empty_image(msname,copyto,channelize=channelize,npix=npix,cellsize=cell);
#    data = ff0[0].data;
#    ff1 = pyfits.open(copyto);
#    ff1[0].data[0:data.shape[0],...] = data;
#    ff1.writeto(copyto,clobber=True);

  # pyrap.images.image() does not appear to like some FITS files we generate (maybe from WSCLEAN), so use
  # CASA to convert them
  casaimage = II("$image.img")
  rm_fr(casaimage)
  im.argo.fits2casa(image,casaimage)
  
  # convert to CASA image
  # see discussion in https://github.com/ska-sa/pyxis/issues/61 -- make chunks in frequency if asked to
  import pyrap.images
  info("image is $casaimage")
  img = pyrap.images.image(casaimage)
  imgshp = img.shape()
  # default chunk list is entire chanel range. Update this if needed
  chunklist = [ (ms.CHANSTART,ms.NUMCHANS,None,None) ]
  if len(imgshp) == 4 and imgshp[0] > 1:
      nimgchan = imgshp[0]
      info("image cube has $nimgchan channels, MS has ${ms.NUMCHANS} channels")
      imgchansize = imgshp[1]*imgshp[2]*imgshp[3]*4  # size of an image channel in bytes
      if chanchunk is None:
          mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')  # e.g. 4015976448
          chanchunk = max((mem_bytes/20)/imgchansize,1)
          info("based on available memory ($mem_bytes), max image chunk is $chanchunk channels")
      if chanchunk < nimgchan:
          mschanstep = ms.NUMCHANS*ms.CHANSTEP/nimgchan
          if ms.NUMCHANS%nimgchan:
              warn("MS channels not evenly divisible into $nimgchan image channels, chunking may be incorrect")
          chunklist = []
          for chan0 in range(0,nimgchan,chanchunk):
              imch0, imch1 = chan0, (min(chan0+chanchunk, nimgchan)-1)
              msch0 = ms.CHANSTART + imch0*mschanstep
              msnch = (imch1-imch0+1)*mschanstep/ms.CHANSTEP
              # overlap each chunk from 1 onwards by a half-chunk back to take care of extrapolated visibilties
              # from previous channel
              if imch0:
                  imch0 -= 1
                  msch0 -= mschanstep/2
                  msnch += mschanstep/2
              info("image chunk $imch0~$imch1 corresponds to MS chunk %d~%d"%(msch0,msch0+msnch-1))
              chunklist.append((msch0, msnch, imch0, imch1));
              
  # even in fill-model mode where it claims to ignore image parameters, the image channelization
  # arguments need to be "just so" as per below, otherwise it gives a GridFT: weights all zero message
  kw0.update(ms=msname, model=casaimage, 
      niter=0, fixed=1, mode="channel", operation="csclean",
      img_nchan=1,img_chanstart=ms.CHANSTART, img_chanstep=ms.NUMCHANS*ms.CHANSTEP)
  if LWIMAGER_VERSION[0] >= 1003001:
      kw0['fillmodel'] = 1;
                   
  blc = [0]*len(imgshp)
  trc = [ x-1 for x in imgshp ]
  # now loop over image frequency chunks
  for ichunk, (mschanstart, msnumchans, imgch0, imgch1) in enumerate(chunklist):
      if len(chunklist) > 1:
          blc[0], trc[0] = imgch0, imgch1
          info("writing CASA image for slice $blc $trc")
          casaimage1 = II("$image.$ichunk.img")
          rm_fr(casaimage1)
          info("writing CASA image for slice $blc $trc to $casaimage1")
          img.subimage(blc,trc,dropdegenerate=False).saveas(casaimage1)
          kw0.update(model=casaimage1)
      else:
          img.unlock()
      # setup imager options
      kw0.update(chanstart=mschanstart, chanstep=ms.CHANSTEP, nchan=msnumchans)
      info("predicting visibilities into MODEL_DATA");
      _run(**kw0);
      if len(chunklist) > 1 and not PREDICT_KEEPIMAGES:
        rm_fr(casaimage1)   
  if not PREDICT_KEEPIMAGES:
      rm_fr(casaimage)
  
  if column != "MODEL_DATA":
    ms.copycol(msname=msname,fromcol="MODEL_DATA",tocol=column);

document_globals(predict_vis,"PREDICT_* MS im.MODEL_IMAGE COPY_IMAGE_TO ms.IFRS ms.DDID ms.FIELD ms.CHANRANGE");      

def make_psf (msname="$MS",**kw):
  """Makes an image of the PSF. All other arguments as per make_image()."""
  make_image(msname,dirty=False,psf=True,**kw);



"""Pyxis module for MS-related operations""";
from Pyxis.ModSupport import *

import os
import sys
import numpy
import math
import argo
import im

# Standard options
gain = 0.1
sigmalevel = 3.0

define('MORESANE_PATH_Template','${im.MORESANE_PATH}','Path to PyMORESANE')

_moresane_args = {'outputname': None,
                  'modelname': None,
                  'residualname': None,
                  'restoredname':None,
                  'singlerun': False,
                  'subregion': None,
                  'scalecount': None,
                  'startscale': 1,
                  'stopscale': 20,
                  'sigmalevel': 4,
                  'loopgain': 0.2,
                  'tolerance': 0.75,
                  'accuracy': 1e-6,
                  'majorloopmiter': 100,
                  'minorloopmiter': 50,
                  'allongpu': False,
                  'decommode': 'ser',
                  'corecount': 1,
                  'convdevice': 'cpu',
                  'convmode': 'circular',
                  'extractionmode': 'cpu',
                  'enforcepositivity': False,
                  'edgesupression': False,
                  'edgeoffset': 0,
                  'mfs':False,
                  'mfs-chanrange':None,
                  'spi-sigmalevel':10,
                  'spec-curv':False,
                  'fluxthreshold': 0,
                  'mask': None
}

def deconv(dirty_image,psf_image,
                 model_image='${im.MODEL_IMAGE}',
                 residual_image='${im.RESIDUAL_IMAGE}',
                 restored_image='${im.RESTORED_IMAGE}',
                 image_prefix='${im.BASENAME_IMAGE}',
                 path='$MORESANE_PATH',**kw):
    """ Runs PyMORESANE """

    # Check if PyMORESANE is where it is said to be
    model_image,residual_image,restored_image,path = interpolate_locals('model_image residual_image restored_image path')
    found_path = argo.findImager(path,imager_name='PyMORESANE') 
    if not found_path:
        abort('could not find PyMORESANE at $path')
   
    kw['modelname'] = model_image
    kw['residualname'] = residual_image
    kw['restoredname'] = restored_image
    
    # make dict of imager arguments that have been specified globally or locally
    args = dict([ (arg,globals()[arg]) for arg in _moresane_args if arg in globals() and globals()[arg] is not None ]);
    args.update([ (arg,kw[arg]) for arg in _moresane_args if arg in kw ])

    x.sh(argo.gen_run_cmd(path,args,suf='--',assign='=',pos_args=[dirty_image,psf_image,image_prefix]))

document_globals(deconv,'im*_IMAGE MORESANE_PATH im.MORESANE_PATH')



"""mqt: Pyxides module for MeqTrees-related functionality (running MeqTrees scripts, etc.)"""

import os.path

# register ourselves with Pyxis
from Pyxis.ModSupport import *
register_pyxis_module();

## find the Cattery
import Timba
import os.path
import ms

_cattery_path = Timba.packages()['Cattery'][0]
sys.path.append(_cattery_path);
if v("ADD_PYXIDES_PATH",True):
  path = os.path.join(_cattery_path,"Pyxides");
  verbose(2,"adding %s to import path. Set ADD_PYXIDES_PATH=False to disable"%path);
  sys.path.append(path);

def_global('CATTERY',_cattery_path,"default path to Cattery scripts");

## default multithread setting
def_global('MULTITHREAD',2,"max number of meqserver threads");

## default TDL config file

## extra TDL options applied to all scripts
def_global('EXTRA_TDLOPTS',"","extra options passed to all TDL scripts");

## pipeliner tool
pipeliner = x.time.args("meqtree-pipeliner.py");
pipeliner_mprof = x.mprof.args("run --python meqtree-pipeliner.py");
def_global("SCRIPT",None,"default TDL script");
def_global("JOB",None,"default TDL job to run");
def_global("SECTION",None,"default section to use in TDL config file");
def_global("TDLCONFIG","tdlconf.profiles","default TDL config file",config=True);
def_global("SAVECONFIG",None,"save effective config to file[:section]")

def_global("MEMPROF",False,"if True, runs python memory profiling");

def run (script="$SCRIPT",job="$JOB",config="$TDLCONFIG",section="$SECTION",saveconfig="$SAVECONFIG",args=[],options={}):
  """Uses meqtree-pipeliner to compile the specified MeqTrees 'script', using 'config' file and config 'section',
  then runs the specified 'job'. 

  If you want to run a script on a given MS, then see mqt.msrun() for a convenient alternative.

  Use a list of 'args' to pass extra arguments to meqtree-pipeliner. Use a dict of 'options' to
  pass extra arguments as key=value.""";
  script,job,config,section,saveconfig = interpolate_locals("script job config section saveconfig");
  section = section or os.path.splitext(os.path.basename(script))[0];
  args = [ "-c $config [$section]" ] + list(args)
  if MULTITHREAD > 1:
      args = [ "--mt $MULTITHREAD" ] + args
  if saveconfig:
      args = [ "--save-config $saveconfig"  ] + args
  if isinstance(options, list):
    for dictionary in options:
      args += [ "%s=%s"%(a,b) for a,b in dictionary.items() ] + \
          [ "$EXTRA_TDLOPTS $script =$job "]
  else:
    args += [ "%s=%s"%(a,b) for a,b in options.items() ] + \
        [ "$EXTRA_TDLOPTS $script =$job" ]; 

 # run pipeliner
  if MEMPROF:
    args.append("--memprof");
    pipeliner_mprof(*args);
  else:
    pipeliner(*args);

document_globals(run,"MULTITHREAD EXTRA_TDLOPTS SCRIPT JOB SECTION TDLCONFIG SAVECONFIG");


def msrun (script="$SCRIPT",job="$JOB",section="$SECTION",config="$TDLCONFIG",args=[],options={}):
  """Like run(), but automatically adds TDL options for the currently selected MS/channels/etc
  (according to what is defined in the 'ms' Pyxides module).""";
  script,job,config,section = interpolate_locals("script job config section");
  return run(script=script,job=job,config=config,section=section,
    args = [ """${ms.MS_TDL} ${ms.CHAN_TDL} ms_sel.ms_ifr_subset_str=${ms.IFRS}""" ] + list(args),
    options=options); 

document_globals(msrun,"MULTITHREAD EXTRA_TDLOPTS SCRIPT JOB SECTION TDLCONFIG");

MSSIM_SCRIPT_Template = "$CATTERY/Siamese/turbo-sim.py" 

def mssim (section="$SECTION",config="$TDLCONFIG",script="$MSSIM_SCRIPT",job="simulate",
           column=None,args=[],options={}):
  """Like runms(), but by default runs the turb-sim.py simulations script from the Cattery.
  Common usage is mssim("simulation_section")
  """;
  script,job,config,section,column = interpolate_locals("script job config section column");
  args = [ """${ms.MS_TDL} ${ms.CHAN_TDL} ms_sel.ms_ifr_subset_str=${ms.IFRS}""" ] + list(args)
  if column is not None:
    args += [ II("ms_sel.output_column=$column") ] 
  return msrun(script=script,job=job,config=config,section=section,
    args=args, options=options); 

document_globals(mssim,"MULTITHREAD EXTRA_TDLOPTS SECTION TDLCONFIG MSSIM*");



"""Pyxis.Internals: various internal machinery of Pyxis"""

import glob
import traceback
import subprocess
import re
import string
import os
import os.path
import inspect
import sys
import fnmatch
import shutil
import shlex

import Pyxis

_preset_doc = """The following global variables control Pyxis behaviour. These may be set from the command line 
as VAR=VALUE, or specified in Pyxis recipes or config files or recipes (as v.VAR = VALUE).

LOG: all output will be logged to this file (except for level-0 status messages, which are duplicated
to the console). Can also be specified as a template (e.g. LOG_Template={$MS:BASE}.log) to make the log
dependent on other variables.

LOG_DISABLE: if True, logfiles will be disabled, and all output will go to the console

LOG_FLUSH: if True, logfiles will be overwritten by each new Pyxis run. Default is to append to existing logfiles.

LOG_HEADER: header string placed at start of each new log file or log session.

VERBOSE: Pyxis verbosity level, default is 0. Higher is more verbose.

OUTDIR: output directory, for use in recipes and config files (note that Pyxis itself in no way enforces
any output to this directory -- it is up to recipes and configs to use $OUTDIR consistently in their naming
schemes filenames).

PYXIS_LOAD_CONFIG: pre-load recipes (pyxis-*.py) and config (pyxis-*.conf) files from current directory when
Pyxis is initialized. Default is true.

PYXIS_AUTO_IMPORT_MODULES: automatically import (into the global namespace) all top-level Pyxides modules 
directly or indirectly invoked by the pre-loaded recipes. Useful for interactive sessions. Default is True.

JOBS: split out up to this many subprocesses to work in parallel, when executing per() commands. 
Default is 1.

JOB_STAGGER: stagger launch of subprocesses by this many seconds. Can be useful to e.g. de-syncronize
disk access in subprocesses. Default is 10.

PERSIST: if False, then per() commands (such as per_ms) will abort processing on any error. If True,
per commands will carry on with other items in the list, and will only report the error afterwards.
"""

# set of protected variables -- assignments to these via templates or assign() will be ignored
_protected_variables = set();

# printed to verbose() after startup
_verbose_startup_message = None; 

def init (context):
  """init internals, attach to the given context""";
  global _debug
  global _info
  global _abort
  global _verbose
  global _warn
  from Pyxis.Commands import _debug,_info,_abort,_verbose,_warn
  # set default output dir
  context.setdefault("OUTDIR",".");
  context.setdefault("JOBS",0);
  context.setdefault("JOB_STAGGER",10);
  context.setdefault("PERSIST",0);
  context.setdefault("PYXIS_LOAD_CONFIG",True);
  context.setdefault("PYXIS_AUTO_IMPORT_MODULES",True);
  # set default verbosity to 1
  preset_verbosity = context.get("VERBOSE",None);
  context.setdefault("VERBOSE",1);
  # set context and init stuff
  Pyxis.Context = context;
  Pyxis._predefined_names = set(context.keys());
  Pyxis.Commands._init(context);
  # loaded modules
  global _namespaces,_superglobals,_modules;
  _namespaces = dict();
  _superglobals = dict();
  _modules = dict();
  # The "v" and "" namespaces correspond to the global context
  # All variables of that context are superglobals.
  _namespaces['v'] = context;
  _superglobals[id(context)] = context;
  # report verbosity
  global _verbose_startup_message;
  if preset_verbosity is None and context['VERBOSE'] == 1:
    _verbose_startup_message = 1,"VERBOSE=1 by default";
  else:
    _verbose_startup_message = 1,"VERBOSE=%d"%context['VERBOSE'];
  # import matplotlib in fork, disable output if fails
  try:
    import matplotlib
  except:
  # no matplotlib? Carry on without
    return;
  # Use the 'agg' backend if X is disabled
  if context.get("NO_X11") or os.system('xdpyinfo >/dev/null 2>/dev/null'):
    matplotlib.use('agg');

def _int_or_str (x):
  """helper function: converts argument to int if possible, else returns string""";
  try:
    return int(x);
  except:
    return str(x);
    
def interpolate_args (args,kws,frame,convert_lists=False): 
  """Helper function to interpolate argument list and keywords using the local dictionary, plus Pyxis.Context globals""";
  return [ interpolate(arg,frame,convert_lists=convert_lists) for arg in args ], \
         dict([ (kw,interpolate(arg,frame,convert_lists=convert_lists)) for kw,arg in kws.items() ]);
  
class ShellExecutor (object):    
  """This is a ShellExecutor object, which is associated with a particular shell command, and can be
used to run that command. The command (and any fixed arguments) are specified in the constructor of 
the object. ShellExecutors are typically created via the Pyxis x, xo or xz built-ins, e.g. as 
      ls = x.ls 
      lsl = x.ls.args("-l")
      ls()             # runs "ls"
      dir="./test"
      lsl("$dir")      # runs "ls -l ./test"
      ls(x=1)          # runs "ls x=1"
  """;
  
  def __init__ (self,name,path,frame,allow_fail=False,get_output=False,bg=False,verbose=1,args0=(),kws0={},args1=(),kws1={},doc=False):
    self.name,self.path = name,path;
    self.allow_fail = allow_fail;
    self.bg  = bg;
    self.argframe = frame;
    self.get_output = get_output;
    self.verbose = verbose;
    self._pre_args,self._pre_kws = (list(args0),kws0);
    self._post_args,self._post_kws = (list(args1),kws1);
    self.__doc__ = "(Shell command: %s)"%str(self);
    if doc is not None:
      self.doc(doc);
      
  def doc (self,doc,append=True):
    """Sets doc string""";
    if append:
      self.__doc__ += "\n%s"%doc;
    else:
      self.__doc__ = doc;
    
  def args (self,*args,**kws):
    """Creates instance of executor with additional args. Local variables of caller are interpolated."""
    args0,kws0 = interpolate_args(self._pre_args,self._pre_kws,self.argframe);
    args1,kws1 = interpolate_args(self._post_args,self._post_kws,self.argframe);
    before = kws.pop("before",None);
    after = kws.pop("after",None);
    kws0.update(kws);
    if before is not None:
      args0 += [ before ] if isinstance(before,str) else list(before);
    if after is not None:
      args1 += [ after ] if isinstance(after,str) else list(after);
    return ShellExecutor(self.name,self.path,inspect.currentframe().f_back,self.allow_fail,
        self.get_output,self.bg,self.verbose,(args0+list(args)),kws0,args1,kws1);

  def __str__ (self):
    return " ".join([self.path or ""]+self._pre_args+["%s=%s"%(a,b) for a,b in self._pre_kws.items()]);
    
  def __repr__ (self):
    return "ShellExecutor: %s"%str(self);
    
  def __call__ (self,*args,**kws):
    """Runs the associated shell command, with additional supplied arguments. Normal arguments are simply converted
    to strings. Keywords are converted to key=value arguments. Local variables of the caller are interpolated."""
    if self.path is None:
      if self.allow_fail:
        _abort("PYXIS: shell command '%s' not found"%self.name);
      _warn("PYXIS: shell command '%s' not found"%self.name);
    else:
      args0,kws0 = interpolate_args(self._pre_args,self._pre_kws,self.argframe,convert_lists=True);
      args1,kws1 = interpolate_args(self._post_args,self._post_kws,self.argframe,convert_lists=True);
      args,kws = interpolate_args(args,kws,inspect.currentframe().f_back,convert_lists=True);
      kws0.update(kws);
      return _call_exec(self.path,get_output=self.get_output,allow_fail=self.allow_fail,bg=self.bg,verbose=self.verbose,
        args=args0+args+args1,kws1=kws1,**kws0);

class ShellExecutorFactory (object):
  """This object can be used to create proxies for shell commands called ShellExecutors."""
  def __init__ (self,allow_fail=False,bg=False,get_output=False,verbose=0):
    self.allow_fail = allow_fail;
    self.bg = bg;
    self.get_output = get_output;
    self.verbose = verbose;
    self.doc_proto = """The '%(name)s' built-in provides an interface for shell commands. Invoking
%(name)s.sh("command") runs a command directly; invoking %(name)s.command returns a ShellExecutor for  
"command", i.e. a Python object that can used to run the command later. For example:
    ls = %(name)s.ls                   # creates a proxy for the ls command
    ls()                        # executes ls
    ls("-l")                    # executes ls -l
    ls = %(name)s("ls")                # alternative syntax for creating a ShellExecutor.\n""";
    
  def __getattr__ (self,command,default=None):
    """Creates a ShellExecutor for a given shell command. For example,
    ls = x.ls     # the ls object is now a ShellExecutor for the ls command
    ls()          # executes ls
    ls("-l")      # executes ls -l
 """
    command = interpolate(command,inspect.currentframe().f_back);
    if command.find('/') >= 0:
      path = command if os.access(command,os.X_OK) else None;
    else:
      path = find_exec(command);
    return ShellExecutor(command,path,None,self.allow_fail,self.get_output,self.bg,self.verbose);
    
  def __call__ (self,*args,**kws):
    """An alternative way to make ShellExecutors, e.g. as x("command arg1 arg2").
    Useful when the command contains e.g. dots or slashes, thus making the x.command syntax unsuitable."""
#    args,kws = interpolate_args(args,kws,inspect.currentframe().f_back);
    if len(args) < 1:
      _abort("can't call %s without arguments"%self.__name__);
    elif len(args) == 1:
      args = args[0].split(" ");
    return ShellExecutor(args[0],args[0],None,allow_fail=self.allow_fail,bg=self.bg,
                verbose=self.verbose,get_output=self.get_output,args0=args[1:],kws0=kws);
    
  def sh (self,*args,**kws):
    """Directly invokes the shell with a command and arguments"""
    commands,kws = interpolate_args(args,kws,inspect.currentframe().f_back,convert_lists=True);
    # run command
    _verbose(self.verbose,"executing '%s':"%(" ".join(commands)));
    flush_log();
    #
    quiet = kws.pop("quiet",False)
    # if stdout/stderr is not a file (as is the case under ipython notebook, then
    # subprocess.Popen() fails. Therefore, in these cases, or if get_output is true, we
    # pipe the output into here via communicate()
    stdout = subprocess.PIPE if self.get_output or type(sys.stdout) is not file else sys.stdout;
    stderr = subprocess.PIPE if type(sys.stderr) is not file else sys.stderr;
    po = subprocess.Popen(["/bin/bash","-c"]+list(commands), preexec_fn=_on_parent_exit('SIGTERM'),
        shell=False,stdout=stdout,stderr=stderr);
    # if piping either output stream, capture it here
    if stdout is subprocess.PIPE or stderr is subprocess.PIPE:
      output,err_output = po.communicate();
      if not self.get_output and output is not None and not quiet:
        sys.stdout.write(output);
        output = None;
      if err_output is not None and not quiet:
        sys.stderr.write(err_output);
    else:
      po.wait();
      output = po.returncode;
    if po.returncode:
      if self.allow_fail:
        _warn("PYXIS: '%s' returns error code %d"%(commands[0],po.returncode));
      else:
        _abort("PYXIS: '%s' returns error code %d"%(commands[0],po.returncode));
    else:
      _verbose(self.verbose+1,"'%s' succeeded"%commands[0]);
    return output;
      
  def __repr__ (self):
    name = self.__name__;
    return "Pyxis built-in %s: access to shell commands. Use help(%s) for details."%(name,name);

class _DictAccessor (object):
  """Helper class that maps dicts to attributes""";
  def __init__ (self,namespace):
    object.__setattr__(self,'namespace',namespace);
    
  def __call__ (self,name,default=""):
    if isinstance(default,str):
      default = interpolate(default,inspect.currentframe().f_back);
    return object.__getattribute__(self,'namespace').get(name,default);
    
  def __getattr__ (self,name,default=""):
    if isinstance(default,str):
      default = interpolate(default,inspect.currentframe().f_back);
    return object.__getattribute__(self,'namespace').get(name,default);
    
  def __setattr__ (self,name,value):
    object.__getattribute__(self,'namespace')[name] = value;
    
  def __contains__ (self,name):
    return name in object.__getattribute__(self,'namespace');

class GlobalVariableSpace (_DictAccessor):
  """The "v" object provides access to global Pyxis variables. Global variables assigned with
  v.VARNAME=value are propagated across all imported Pyxis modules. They can then be referenced
  directly as VARNAME. The "v" object itself provides a number of extra features:

  v.VARNAME evaluates to the variable VARNAME, or to the empty string, if VARNAME is not defined
  v('VARNAME',default) evaluates to the variable VARNAME, or to default if VARNAME is not defined.
     If default is a string, local variables of the caller are interpolated.
  v.VARNAME=value assigns to a global variable, and also causes templates to be re-evaluated, and 
    other implicit variable-related actions to be taken. In particular, v.LOG="logfile" will set a 
    new log destination.
    
  v.define('VARNAME',value,doctext): assigns to variable VARNAME, and sets a documentation string
  v.doc('VARNAME'): returns the documentation string for a variable, or '' if none
  v.doc('VARNAME',doctext): sets the documentation string for a variable
  """;
  
  def __init__ (self,namespace):
#    object.__setattr__(self,'_docs',{});
    _DictAccessor.__init__(self,namespace);
    
  def __setattr__ (self,attr,value):
    assign(attr,value,namespace=self.namespace,frame=inspect.currentframe().f_back); 
    
  def __repr__ (self):
    name = object.__getattribute__(self,'__name__');
    return "Pyxis built-in %s: global variable space. Try %s.VARNAME, or help(%s)."%(name,name,name);
  
  def define (self,name,defvalue,doc=None):
    """Defines a variable, with an optional documentation string.""";
    frame = inspect.currentframe().f_back;
    register_superglobal(frame,name if not name.endswith("_Template") else name[:-len("_Template")]);
    ns = object.__getattribute__(self,'namespace');
    if name not in ns:
      assign(name,defvalue,namespace=ns,frame=frame); 
    if doc:
      ns.setdefault('_symdocs',{})[name] = doc;
  
  def doc (self,name,text=None):
    """doc(name) gets the documentation string for a variable, or '' if none is set.
    doc(name,text) sets the documentation string for the variable.""";
    ns = object.__getattribute__(self,'namespace');
    docs = ns.get('_symdocs',{});
    if text is None:
      return docs.get(name,'');
    docs[name] = text;
    return text;
    
class ShellVariableSpace (_DictAccessor):
  """This object provides quick access to environment (i.e. shell) variables. Use e.g.
  E.HOME or E("HOME",default) to access a shell variable. If default is a string, local 
  variables of the caller are interpolated.
  """;
  def __init__ (self):
    _DictAccessor.__init__(self,os.environ);
    
  def __repr__ (self):
    name = object.__getattribute__(self,'__name__');
    return "Pyxis built-in %s: quick access to shell variables. Try %s.VARNAME, or help(%s)."%(name,name,name);

def interpolate (arg,frame,depth=1,ignore=set(),skip=set(),convert_lists=False):
  """Interpolates strings: substitutes $var and ${var} with the corresponding variable value from 
  (in order of lookup):
  
  * the locals and globals of the given frame (must be a frame object: see inspect module for details)
  * the locals and globals of outer frames to a depth of 'depth' (if >1)
  * the global Pyxis context. 
  
  Alternatively, if frame is a dict, then lookup happens in frame, then the global Pyxis context.
  
  If arg is a string, does interpolation and returns new string.
  
  If arg is a dict, does interpolation on every string-type key in the dict (except for those in 'skip'), using the 
    dict itself as a source of symbols (plus the global variables). Returns the dict.
    
  If set, 'ignore' is a container of symbols which will interpolate to an empty string.
  """;
  # setup lookup dictionaries based on frame and depth
  if isinstance(frame,dict):
    lookups = [ frame,Pyxis.Context ];
  else:
    lookups = [];
    while depth>=0 and frame:
      lookups += [ frame.f_locals,frame.f_globals ];
      frame = frame.f_back
      depth -= 1
    lookups.append(Pyxis.Context);
  # convert lists to strings
  if isinstance(arg,(list,tuple)) and convert_lists:
    arg = ",".join(map(str,arg));
  # interpolate either a single string, or a dict recursively
  if isinstance(arg,dict):
    arg,arg0 = arg.copy(),arg;
    if arg0 is not lookups[0]:
      lookups = [arg] + lookups;
    defdict = DictProxy(lookups,ignore);
  # interpolate until things stop changing, but quit after 20 loops
    for count in range(20):
      updates = {};
      for key,value in arg.items():
        # interpolate string variables
        if key not in skip and isinstance(value,str):
          defdict.ignores = [value];
          newvalue = SmartTemplate(value).safe_substitute(defdict);
#          print "%s: %s->%s"%(key,value,newvalue);
          if newvalue != value:
            updates[key] = newvalue;
      # apply updates, unless things stop
#      print updates;
      if not updates:
        break;
      arg.update(updates);
    return arg;
  # strings are interpolated
  elif isinstance(arg,str):
    defdict = DictProxy(lookups,ignore);
    return SmartTemplate(str(arg)%defdict).safe_substitute(defdict);
  # all other types returned as-is
  else:
    return arg;

# RE pattern matching the [PREFIX<][NAMESPACES.]NAME[?DEFAULT][:BASE|DIR|FILE|BASEPATH][>SUFFIX] syntax
_substpattern = \
  "(?i)((?P<prefix>[^{}]+)<)?(?P<name>[._a-z][._a-z0-9]*)(\\?(?P<defval>[^}\\$]*?))?(:(?P<command>BASE|DIR|FILE|BASEPATH))?(>(?P<suffix>[^{}]+))?"
    
class SmartTemplate (string.Template):
  pattern = "(?P<escaped>\\$\\$)|(\\$(?P<named>[_a-z][_a-z0-9]*))|(\\${(?P<braced>%s)})|(?P<invalid>\\$)"%_substpattern;

class DictProxy (object):
  itempattern = re.compile(_substpattern+"$");
  
  def __init__ (self,dicts,ignores):
    self.dicts = dicts;
    self.ignores = frozenset(ignores);
    
  def __getitem__ (self,item):
    # parse the item as a [PREFIX<]NAME[?DEFAULT][:BASE][>suffix] combo
    match = DictProxy.itempattern.match(item);
    if not match:
      return ""#,item;
    prefix,name,defval,command,suffix = match.group("prefix","name","defval","command","suffix");
    if name in self.ignores:
      return "";
    # is there an explicit namespace? otherwise use the default "merged" one
    if '.' in name:
      namespace,name = name.rsplit(".",1);
      namespace = _namespaces.get(namespace);
      if namespace is None:
        return ""; #,item;
      value = namespace.get(name,defval);
    else:
      for dd in self.dicts:
        value = dd.get(name);
        if value is not None:
          break;
    if value is None or value == '':
      value = defval;
    # check for commands
    if isinstance(value,str) and command:
      # for dir/file.ext, returns "file"
      if command.upper() == "BASE":
        value = value and os.path.basename(value);
        value = value and os.path.splitext(value)[0];
      # for dir/file.ext, returns "dir/file"
      elif command.upper() == "BASEPATH":
        value = value and os.path.splitext(value)[0];
      # for dir/file.ext, returns "dir"
      elif command.upper() == "DIR":
        value = (value and os.path.dirname(value)) or ".";
      # for dir/file.ext, returns "file.ext"
      elif command.upper() == "FILE":
        value = value and os.path.basename(value);
    return (prefix or "")+str(value)+(suffix or "") if value not in ('',None) else "";
    
  def __contains__ (self,item):
    return True;

def _resolve_namespace (name,frame,default_namespace=None,autoimport=False):
  if '.' in name:
    nsname,name = name.rsplit(".",1);
    if autoimport:
      _autoimport(nsname);
    namespace = _namespaces.get(nsname);
    if namespace is None:
      raise ValueError("invalid namespace %s"%nsname);
  else:
    namespace = default_namespace or frame.f_globals;
  return namespace,name;
   
def assign (name,value,namespace=None,default_namespace=None,interpolate=True,frame=None,append=False,autoimport=False,verbose_level=2):
  """Assigns value to variable, then reevaluates templates etc."""
  frame = frame or inspect.currentframe().f_back;
  # find namespace
  if not namespace:
    namespace,name = _resolve_namespace(name,frame,default_namespace,autoimport=autoimport);
  modname = namespace.get('__name__',"???") if namespace is not Pyxis.Context else "v";
    # skip protected variables in global context
  if name in namespace.setdefault('__pyxis_protected_variables',set()):
    _verbose(verbose_level,"ignoring assign('%s.%s',...): protected variable"%(modname,name));
    return;
  # interpolate if asked to, unless this is a template, which are never interpolated
  if interpolate and not name.endswith("_Template"):
    value1 = Pyxis.Internals.interpolate(value,frame);
  else:
    value1 = value;
  # append if asked to
  if append:
    value0 = str(namespace.get(name,""));
    if value0:
      value1 = "%s %s"%(value0,value1);
  # get list of namespaces in which to set variable
  namespaces = [ namespace ];
#  namespace[name] = value1;
#  _verbose(verbose_level,"setting %s.%s=%s"%(modname,name,value1));
  # get superglobals associated with this namespace: if the variable is one of them, then propagate the new value
  # across all other namespaces using that superglobal
  superglobs = _superglobals.get(id(namespace),[]);
  if name in superglobs:
    namespaces += [ ns for ns in _namespaces.values() if name in _superglobals.get(id(ns)) and ns is not namespace ];
  # now assign
  for ns in namespaces:
    nsname = ns['__name__'] if ns is not Pyxis.Context else "v";
    _verbose(verbose_level,"setting %s.%s=%s"%(nsname,name,value1));
    ns[name] = value1;
    # if assigning a template, make sure the template is re-enabled
    if name.endswith("_Template"):
      ns.get('__pyxis_template_ids',{}).pop(name[:-len("_Template")],None);
  # reprocess templates
  assign_templates();

def unset (name,namespace=None,frame=None,verbose_level=2):
  """Unsets variable. Reactivates any templates associated with variable."""
  frame = frame or inspect.currentframe().f_back;
  # find namespace
  if not namespace:
    namespace,name = _resolve_namespace(name,frame=frame);
  # get list of namespaces from which to unset
  namespaces = [ namespace ];
  # get superglobals associated with this namespace: if the variable is one of them, then propagate the un-setting
  # across all other namespaces using that superglobal
  superglobs = _superglobals.get(id(namespace),[]);
  if name in superglobs:
    namespaces += [ ns for ns in _namespaces.values() if name in _superglobals.get(id(ns)) and ns is not namespace ];
  # now loop over all namespaces in which to unset
  for ns in namespaces:
    modname = ns.get('__name__',"???") if ns is not Pyxis.Context else "v";
    if name in ns:
      _verbose(verbose_level,"unsetting %s.%s"%(modname,name,value1));
      ns.pop(name);
    tmpldict = ns.get('__pyxis_template_ids',{});
    if name in tmpldict:
      tmpldict.pop(name);
      _verbose(verbose_level,"  and re-enabling associated template %s.%s_Template"%(modname,name));
  # reprocess templates
  assign_templates();


class DependsHandler (object):
  _handlers = {};
  
  def __init__ (self,varname,vardepend,frame=None):
    # resolve namespaces
    frame = frame or inspect.currentframe().f_back;
    self.namespace,self.name = _resolve_namespace(varname,frame);
    self.depnamespace,self.depname = _resolve_namespace(vardepend,frame);
    # add to list
    self.patterns = [];
    DependsHandler._handlers.setdefault((id(self.depnamespace),self.depname),[]).append(self);
    
  def __call__ (self,pattern,value):
    self.patterns.append((pattern,value));
    return self;
    
  def default (self,value):
    self.defval = value;
    return self;
    
  @staticmethod
  def update (namespace,name,value):
    result = False;
    for hdl in DependsHandler._handlers.get((id(namespace),name),[]):
      result |= hdl.check(value);
    return result;
    
_in_assign_templates = False;

def assign_templates ():
  """For every variable in a Pyxis module (or the global context) that ends with "_Template", assigns value 
  to it by interpolating the template.""";
  ## non-reentrant, otherwise any call to Pyxis methods from within a template is liable to cause recursion
  global _in_assign_templates;
  if _in_assign_templates:
    return;
  _in_assign_templates = True;
  for count in range(100):
    updated = False;
    for modname,context in list(_namespaces.items()):
      superglobs = _superglobals[id(context)];
      newvalues = {};
      templdict = context.setdefault("__pyxis_template_ids",{});
      # interpolate new values for each variable that has a _Template equivalent
      for var,value in list(context.items()):
        if var.endswith("_Template"):
          varname = var[:-len("_Template")];
          # skip protected variables in global context
          if varname in context.setdefault('__pyxis_protected_variables',set()):
            _verbose(3,"ignoring template assignment of %s.%s: protected variable"%(modname,varname));
            continue;
          # get old value of variable
          oldvalue = varvalue = context.get(varname);
          # check if template is defined in the wrong place, superglobal templates must be defined
          # in the superglobal context
          if varname in superglobs and context is not Pyxis.Context:
            _abort("%s.%s defined for superglobal v.%s. Fix your scripts please: use v.%s instead"%
                    (modname,var,varname,var));
          # check if template is still active
          if varname in templdict:
            idvar = templdict[varname];
            # check if template has been disabled by setting __pyxis_template_ids[var] = None
            if idvar is None:
              _verbose(3,"template for %s.%s has been disabled, ignoring"%(modname,var));
              continue;
            # check if variable has been explicitly assigned to since the last time the template
            # got evaluated (i.e. if the id has changed). If so, disable template
            if idvar != id(oldvalue):
              _verbose(3,"value for %s.%s has been explicitly set [%x, was %x], disabling template"%(modname,var,id(oldvalue),idvar));
              templdict[varname] = None;
              continue;
          # catch all template errors below
          try:
            # string templates are simply interpolated 
            if isinstance(value,str):
              varvalue = interpolate(value,context);
            # templates of the form
            # SELECT_EXPR,{pattern:value,pattern:value,...}  [,ELSEVALUE]
            # or SELECT_EXPR,[ (pattern,value),(pattern,value),... ]  [,ELSEVALUE]
            elif isinstance(value,tuple):
              if len(value) not in (2,3) or not isinstance(value[0],str) or not isinstance(value[1],(list,tuple,dict)):
                raise TypeError("invalid select clause");
              select_expr = interpolate(value[0],context);
              # loop through patterns, if one matches the select expression, return that value
              for pair in ( value[1] if not isinstance(value[1],dict) else iter(value[1].items()) ):
                if len(pair) != 2 or not isinstance(pair[0],str):
                  raise TypeError("invalid element in select clause");
                if fnmatch.fnmatch(select_expr,interpolate(pair[0],context)):
                  varvalue = pair[1];
                  break;
              # no patterns match? Look for ELSEVALUE, if defined
              else:
                if len(value) == 3:
                  varvalue = value[2];
                elif isinstance(value[1],dict) and 'default' in value[1]:
                  varvalue = value[1]['default'];
            # list templates are interpolated per-element
            elif isinstance(value,list):
              varvalue = [ interpolate(val) for val in value ];
            # callable templates are called directly
            elif callable(value):
              varvalue = value();
          except:
            traceback.print_exc();
            _warn("PYXIS: error evaluating template %s"%var);
          if varvalue is not oldvalue:
            newvalues[varname] = varvalue;
      # update dict
      for var,value in newvalues.items():
        oldval = context.get(var);
        if oldval != value:
          updated = True;
          # set value, and set id in templdict for later comparison
          context[var] = value;
          templdict[var] = id(value); 
          # if variable is superglobal, propagate it to global context
          if var in superglobs:
            Pyxis.Context[var] = value;
          _verbose(3,"%s templated value %s.%s=%s [%x]"%("initialized" if oldval is None else "updated",modname,var,value,id(value)));
    if not updated:
      break;
  else:
    _abort("Too many template assignment steps. This can be caused by templates that cross-reference each other");
  # propagate superglobals
  for ns in _namespaces.values():
    if ns is not Pyxis.Context:
      for var in _superglobals.get(id(ns)):
        if var in Pyxis.Context:
          ns[var] = Pyxis.Context[var];
  # set logger, in case LOG value has changed
  set_logfile(Pyxis.Context.get('LOG',None));
  _in_assign_templates = False;

_current_logfile = None;
_current_logobj = None;
_warned_nolog = None;
_visited_logfiles = set();

def flush_log ():
  global _current_logobj,_current_logfile;
  _current_logobj and _current_logobj.flush();

def update_log ():
  global _current_logobj,_current_logfile;
  _current_logobj and _current_logobj.seek(0,2);

def get_logfile ():
  global _current_logobj,_current_logfile;
  return _current_logobj,_current_logfile;

def set_logfile (filename,quiet=False):
  """Starts logging to the specified file""";
  global _current_logfile,_current_logobj,_warned_nolog;
  import Pyxis
  if Pyxis.Context.get("LOG_DISABLE"):
    if not _warned_nolog:
      _warn("logfiles disabled, forcing Pyxis output to console and ignoring LOG assignments");
      _warned_nolog = True;
    return;
  import Pyxis.ModSupport
  if filename is not None:
    filename = str(filename);
  if filename == "-" or not filename:
    filename = None;
  if filename != _current_logfile:
    if Pyxis.Context.get('get_ipython'):
      if not _warned_nolog:
        _warn("running inside ipython, forcing Pyxis output to console and ignoring LOG assignments");
        _warned_nolog = True;
      return;
    if not quiet:
      _info("redirecting log output to %s"%(filename or "console"),console=True);
    if filename is None:
      sys.stdout,sys.stderr = sys.__stdout__,sys.__stderr__;
      _current_logobj = None;
    else:
      mode = "a";
      # append to file if name starts with +, or if file has already been used as a log this session, or if flush is off
      if Pyxis.Context.get("LOG_FLUSH"):
        mode = "w";
        if filename[0] == '+':
          filename = filename[1:];
          mode = "a";
        if filename in _visited_logfiles:
          mode = "a";
      Pyxis.ModSupport.makedir(os.path.dirname(filename),no_interpolate=True);
      _current_logobj = sys.stdout = sys.stderr = open(filename,mode);
      hdr = Pyxis.Context.get("LOG_HEADER");
      if filename not in _visited_logfiles and hdr:
        _info(hdr,quiet=True);
        _visited_logfiles.add(filename);
#    if _current_logfile:
#      pass;
#      _info("log continued from %s"%_current_logfile);
#    else:
#      _info("log started");
    _current_logfile = filename;

_initconf_done = False;  
_config_files = [];

def initconf (force=False,files=[],directory="."):
  """Loads configuration from specified files, and/or from default files (in directory, if directory is not None)""";
#  print "initconf",force,Pyxis.Context.get("PYXIS_LOAD_CONFIG",True);
  if not force and not Pyxis.Context.get("PYXIS_LOAD_CONFIG",True):
    return;
  global _initconf_done;
  if not _initconf_done:
    _initconf_done = True;
  if files:
    _verbose(1,"loading config files and scripts:",*files)
  if directory:
    autofiles = glob.glob("%s/pyxis*.py"%directory) + glob.glob("%s/pyxis*.conf"%directory);
    if autofiles:
      _verbose(1,"auto-loading from %s:"%directory,*[ f.rsplit("/",1)[-1] for f in autofiles ]);
      files = list(files) + autofiles;
  global _config_files;
  _config_files = files;
  # remember current set of globals
  oldsyms = frozenset(iter(Pyxis.Context.keys()));
  # load config files -- all variable assignments go into the Pyxis.Context scope
  cwd = os.getcwd();
  try:
    for filename in files:
      Pyxis.Commands.loadconf(filename,inspect.currentframe().f_back,chdir=True);
  finally:
    os.chdir(cwd);
  assign_templates();
  # report on global symbols
  report_symbols("global",[],
      [ (name,obj) for name,obj in Pyxis.Context.items() 
        if name not in oldsyms and not name.startswith("_") and name not in ("In","Out") and name not in Pyxis.Commands.__dict__ ]);
  # make all auto-imported Pyxides modules available to global context
  toplevel = [ m for m in _modules.keys() if not '.' in m and (m in sys.modules or "Pyxides."+m in sys.modules) ];
  if Pyxis.Context.get("PYXIS_AUTO_IMPORT_MODULES",True) and toplevel:
    _verbose(1,"importing top-level modules (%s) for you. Preset PYXIS_AUTO_IMPORT_MODULES=False to disable."%", ".join(toplevel));
    for mod in toplevel:
      Pyxis.Context[mod] = sys.modules.get(mod,sys.modules.get("Pyxides."+m));
  
def loadconf (filename,frame=None,chdir=True):
  """Loads config file""";
  filename = interpolate(filename,frame or inspect.currentframe().f_back);
  _verbose(2,"loading %s"%filename);
  load_package(os.path.splitext(os.path.basename(filename))[0],filename,chdir=chdir);
  
  
def saveconf ():
  """Saves config files to OUTDIR""";
  OUTDIR = Pyxis.Context['OUTDIR'] or '.';
  # make set of all config files
  configs = set(_config_files);
  for m,globs in _namespaces.items():
    for fvar in globs.get('_config_files',[]):
      if fvar in globs:
        configs.add(globs[fvar]);
  # now back them up
  for ff in configs:
    dest = os.path.join(OUTDIR,os.path.basename(ff));
    if os.path.exists(ff):
      if not os.path.exists(dest) or os.path.getmtime(dest) < os.path.getmtime(ff):
        _info("copying %s to %s"%(os.path.basename(ff),OUTDIR));
        shutil.copyfile(ff,dest);
  

def load_package (pkgname,filename,chdir=True,report=True):
  """Loads 'package' file into the Context namespace and reports on new global symbols"""
#  oldstuff = Pyxis.Context.copy();
  try:
    oldpath = list(sys.path);
    dirname = os.path.dirname(filename);
    if dirname not in oldpath:
      sys.path.append(dirname);
    try:
      exec(file(filename),Pyxis.Context);
    finally:
      sys.path = oldpath;
  except SystemExit:
    raise;
  except KeyboardInterrupt:
    raise;
  except:
    traceback.print_exc();
    _abort("PYXIS: error parsing %s, see output above and/or log for details"%filename);
#  newnames =  [ (name,obj) for name,obj in Pyxis.Context.iteritems() 
#                 if not name.startswith("_") and not name in oldstuff ];
#  report_symbols(pkgname,newnames);
  
def register_superglobal (frame,sym):
  """Helper function. Registers the given symbol as a superglobal for the module given by 'frame'.
  If this is not a registered Pyxis module, does nothing.""";
  globs = frame.f_globals;
  modname = globs['__name__'];
  _sgs = _superglobals.get(id(globs));
  if _sgs is not None:
    if sym in _sgs:
      _verbose(3,"%s already registered as a superglobal for module '%s'"%(sym,modname));
      return;
    _verbose(2,"defining superglobal %s in module '%s'"%(sym,modname));
    _sgs.add(sym);
  else:
    _verbose(3,"'%s' is not a registered module, ignoring superglobal '%s'"%(modname,sym));
  
def is_superglobal (globs,sym):
  """Returns True if symbol is registered as a superglobal in the module whose globals are glob"""
  return sym in _superglobals.get(id(globs),{});
  
def report_symbols (pkgname,superglobs,syms):
  if Pyxis.Context['VERBOSE'] >= 2:
    # remove modules from symbols
    syms = [ (name,obj) for name,obj in syms if not inspect.ismodule(obj) ];
    varibs = sorted([name for name,obj in syms if not callable(obj) and not name.endswith("_Template") ]);
    funcs = sorted([name for name,obj in syms if callable(obj) and not name.endswith("_Template") and not isinstance(obj,ShellExecutor) ]);
    shtools = sorted([name for name,obj in syms if callable(obj) and not name.endswith("_Template") and isinstance(obj,ShellExecutor) ]);
    temps = sorted([name[:-9] for name,obj in syms if name.endswith("_Template") ]);
    if funcs:
      _verbose(2,"%s functions:"%pkgname," ".join(funcs));
    if shtools:
      _verbose(2,"%s external tools:"%pkgname," ".join(shtools));
    if superglobs:
      _verbose(2,"%s superglobals:"%pkgname," ".join(superglobs));
    if varibs:
      _verbose(2,"%s variables:"%pkgname," ".join(varibs));
    if temps:
      _verbose(2,"%s templates for:"%pkgname," ".join(temps));
    
def find_exec (cmd):
  """Finds shell executable in PATH"""
  for path in os.environ["PATH"].split(":"):
    filename = os.path.join(path,cmd);
    if os.access(filename,os.X_OK):
      return filename;
  return None;
  
_bg_processes = [];  
  
def _call_exec (path,args,kws1={},**kws):
  """Helper function: calls external program with the given arguments and keywords
  (each kw dict element is turned into a name=value argument)""";
  allow_fail = kws.pop('allow_fail',False);
  bg = kws.pop('bg',False);
  verbose = kws.pop('verbose',1);
  get_output = kws.pop('get_output',None);
  quiet = kws.pop('quiet',None);
  # default is to split each argument at whitespace, but split_args=False passes them as-is
  split = kws.pop('split_args',True);
  # build list of arguments
  args1 = [path];
  if split:
    for arg in args:
      args1 += shlex.split(arg) if isinstance(arg,str) else [ str(arg) ];
    # eliminate empty strings
    args1 = [ x for x in args1 if x ];
  else:
    args1 += list(map(str,args));
  # eliminate empty strings when splitting
  args = args1 + ["%s=%s"%(a,b) for a,b in kws.items()] + ["%s=%s"%(a,b) for a,b in kws1.items()];
  # if command is 'time', then actual command is second argument
  cmdname = args[1] if args[0] == "time" or args[0] == "/usr/bin/time" else args[0];
  # run command
  args = [ x for x in args if x ];
  flush_log();
  if bg:
    global _bg_processes;
    po = subprocess.Popen(args,preexec_fn=_on_parent_exit('SIGTERM'));
    _bg_processes.append(po);
    _verbose(verbose,"executing '%s' in background: pid %d"%(" ".join(args),po.pid));
  else:
    _verbose(verbose,"executing '%s':"%(" ".join(args)));
    stdout,stderr = sys.stdout,sys.stderr;
    if quiet:
      stdout = stderr = file('/dev/null','w')
    # if stdout/stderr is not a file (as is the case under ipython notebook, then
    # subprocess.Popen() fails. Therefore, in these cases, or if get_output is true, we
    # pipe the output into here via communicate()
    if get_output or type(stdout) is not file or type(stderr) is not file:
      stdout = stderr = subprocess.PIPE;
    type(stdout) is file and stdout.flush();
    type(stderr) is file and stderr.flush();
    po = subprocess.Popen(args,preexec_fn=_on_parent_exit('SIGTERM'),
      stdout=stdout,stderr=stderr);
    if stdout is subprocess.PIPE:
      output,err_output = po.communicate();
      if not get_output:
        sys.stdout.write(output);
        output = None;
      sys.stderr.write(err_output);
    else:
      po.wait();
      output = po.returncode;
      err_output = None;
    if po.returncode:
      if allow_fail:
        _warn("%s returned error code %d"%(cmdname,po.returncode));
      else:
        _abort("%s returned error code %d"%(cmdname,po.returncode));
    else:
      _verbose(verbose+1,"%s succeeded"%cmdname);
    return output;


## Function to ensure that child processes are killed when Pyxis is killed, see subprocess.Popen calls above
## source: http://www.evans.io/posts/killing-child-processes-on-parent-exit-prctl/
import signal
from ctypes import cdll

def _on_parent_exit(signame):
    """Return a function to be run in a child process which will trigger
    SIGNAME to be sent when the parent process dies.""";
    signum = getattr(signal, signame)
    # Constant taken from http://linux.die.net/include/linux/prctl.h
    _PR_SET_PDEATHSIG = 1
    def set_parent_exit_signal():
        # http://linux.die.net/man/2/prctl
        result = cdll['libc.so.6'].prctl(_PR_SET_PDEATHSIG, signum)
        if result != 0:
            raise RuntimeError('prctl failed with error code %s' % result)
    return set_parent_exit_signal

def _autoimport (modname):
  if modname not in _namespaces:
    _verbose(1,"auto-importing module %s"%modname);
    Pyxis.Context[modname] = __import__(modname,Pyxis.Context);
    assign_templates();
    
    
_re_mod_command = re.compile("^(\w[\w.]+)\\.(\w+)$");
    
def find_command (comname,frame=None,autoimport=True):
  """Locates command by name. 
  If command is of form "module.command", attempts to resolve it to a callable.
  Will import module if it is not found and autoimport=True.
  Else if command is present (as a callable) in Pyxis.Context, returns that.
  Otherwise checks the path for a binary by that name, and returns a callable to call that.
  Else aborts.""";
  comname = interpolate(comname,frame or inspect.currentframe().f_back);
  # try auto-import
  m = _re_mod_command.match(comname);
  if m and autoimport:
    _autoimport(m.group(1));
  # first, try to evaluate to a callable function
  try:
    comcall = eval(comname,Pyxis.Context);
    if callable(comcall):
      return comcall;
  except:
    pass;
  # failing that, look for a shell command
  if comname[0] == "?":
    comname = comname[1:];
    allow_fail = True;
  else:
    allow_fail = False;
  path = find_exec(comname);
  if path is None:
    _abort("undefined command '%s'"%comname);
  # make callable for this shell command
  return lambda *args:_call_exec(path,allow_fail=allow_fail,args=args);

_re_assign = re.compile("^(\w[\w.]*)(\\+?=)(.*)$");
_re_command1 = re.compile("^(\\??\w[\w.]*)\\[(.*)\\]$");
_re_command2 = re.compile("^(\\??\w[\w.]*)\\((.*)\\)$");
  
def _parse_cmdline_value (value):
  """Helper function to parse the VALUE portion of VAR=VALUE commands.
  Evaluates the value string as a Python expression, using eval() with superglobals.
  If this fails, then returns the value directly as a string.
  
  This implies that command-line arguments will work as follows:
  
  Argument:              Corresponding Python code:
  ---------              --------------------------
  VAR=x                  VAR = "x"    if superglobal x is undefined, else VAR=x
  "VAR='x'"              VAR = "x"    note that shell will swallow the outer quotes
  VAR=:x:                VAR = "x"    leading/trailing colon striped -- makes it easier to pass literal strings from the shell
  VAR=1                  VAR = 1
  VAR=complex(1)         VAR = 1+0j
  "VAR=dict(x='y')"      VAR = dict(x='y')
  VAR=x=1                VAR = "x=1"
  """
  if len(value)>1 and value[0] == ':' and value[-1] == ":":
    return value[1:-1];
  try:
    return eval(value,Pyxis.Context);
  except:
    return value;

def print_doc (symbol):
  import pydoc;
  # if symbol is "module.symbol", try auto-import
  match_module = _re_mod_command.match(symbol);
  if match_module:
    _autoimport(match_module.group(1));
  # try to evaluate the symbol to 'what'
  what = None
  try:
    what = eval(symbol,Pyxis.Context);
  except:
    # if symbol does not contain a dot, try to treat it as a module name
    if '.' not in symbol:
      try:
        _autoimport(symbol)
        what = eval(symbol,Pyxis.Context)
      except:
        pass
  if what is None:
    print("Pyxis doesn't know anything about '%s'"%symbol);
    return;
  docs = pydoc.render_doc(what).split("\n");
  if docs[0].startswith("Python Library"):
    docs[0] = "Pyxis documentation for %s (%s):"%(symbol, type(what).__name__);
  print("\n".join(docs));

def run (*commands):
  """Runs list of commands""";
  import Pyxis.Commands
  assign_templates();
  # _debug("running",commands);
  frame = inspect.currentframe().f_back;
  for step,command in enumerate(commands):
    # if command is callable, call directly
    if not callable(command):
      # interpolate the command
      command = command.strip();
      _verbose(1,"running command %s"%command);
      # syntax 1: VAR=VALUE or VAR+=VALUE
      match = _re_assign.match(command);
      if match:
        name,op,value = match.groups();
        # assign variable -- note that templates are not interpolated
        Pyxis.Commands.assign(name,_parse_cmdline_value(value),frame=frame,append=(op=="+="),autoimport=True);
        continue;
      # syntax 2: command(args) or command[args]. command can have a "?" prefix to make success optional
      match = _re_command1.match(command) or _re_command2.match(command);
      if match:
        comname,comargs = match.groups();
        # split up arguments
        args = [];
        kws = {};
        for arg in re.split(",," if comargs.find(",,") >=0 else ",",comargs):
          arg = interpolate(arg,frame).strip();
          match = re.match("^(\w+)=(.*)$",arg);
          if match:
            kws[match.group(1)] = _parse_cmdline_value(match.group(2));
          else:
            args.append(_parse_cmdline_value(arg));
        # if command is 'help', disable logging
        logfile = None;
        if comname == "help":
#          comname = 'pydoc.render_doc'
          if _current_logobj:
            logfile = _current_logfile;
            set_logfile(None);
        _initconf_done or initconf(force=True);  # make sure config is loaded
        comcall = find_command(comname,inspect.currentframe().f_back,autoimport=True);
        result = comcall(*args,**kws);
        assign_templates();
        if comname == 'pydoc.render_doc':
          print("help:",result);
        # reset logging, if disabled for 'help'
        if logfile:
          set_logfile(logfile);
        continue;
      # syntax 3: standalone command. This better be found!
      _initconf_done or initconf(force=True);  # make sure config is loaded
      comcall = find_command(command,inspect.currentframe().f_back);
    # fall through here if command is callable
    else:
      comcall = command;
      _verbose(1,"running command %s"%command.__name__);
    comcall();
    assign_templates();
  
